#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod

__all__ = [
    "BaseExperiment"
]


class BaseExperiment(ABC):
    """
    Base abstract class for Experiment.
    """

    @abstractmethod
    def runs(self, *, filter: str):
        """Get the historical runs but with WML Pipeline name filter.

        Parameters
        ----------
        filter: str, required
            WML Pipeline name to filter the historical runs.
        """
        pass



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod


class Serializer(ABC):
    """
    Abstract class for Serializer
    """
    @abstractmethod
    def serialize(self, msg):
        """
        Serialize a message

        :param msg: message to serialize
        :type msg: `Message`
        :return: serialized byte stream
        :rtype: `b[]`
        """
        pass

    @abstractmethod
    def deserialize(self, serialized_byte_stream):
        """
        Deserialize a byte stream to a message

        :param serialized_byte_stream: byte stream
        :type serialized_byte_stream: `b[]`
        :return: deserialized message
        :rtype: `Message`
        """
        pass



__all__ = ["ExperimentDataLoader"]

import pandas as pd

# Note: Use torch DataLoader and Dataset if available, if not, use mocked ones.
# SimpleDataLoader is the simplest wrapper ever for such thing, it is not compatible with torch anymore.
try:
    from torch.utils.data import DataLoader, IterableDataset

except ImportError:

    class SimpleDataLoader:
        def __init__(self, dataset, collate_fn):
            self.dataset = dataset
            self.collate_fn = collate_fn

        def __iter__(self):
            return (data for data in self.dataset)

    DataLoader = SimpleDataLoader
    IterableDataset = object
# --- end note


class ExperimentDataLoader(DataLoader):
    """Experiment Data Loader based on torch DataLoader.
    It interacts with the Flight Service to provide data batch stream

    Example
    -------
    >>> experiment_metadata = {
    >>>     "prediction_column": 'species',
    >>>     "prediction_type": "classification",
    >>>     "project_id": os.environ.get('PROJECT_ID'),
    >>>     'wml_credentials': wml_credentials
    >>> }

    >>> connection = DataConnection(data_asset_id='5d99c11a-2060-4ef6-83d5-dc593c6455e2')


    >>> iterable_dataset = ExperimentIterableDataset(connection=connection,
    >>>                                              enable_sampling=False,
    >>>                                              experiment_metadata=experiment_metadata)

    >>> data_loader = ExperimentDataLoader(dataset=iterable_dataset)

    >>> for data in data_loader:
    >>>     print(data)
    """

    def __init__(self, dataset: "IterableDataset"):
        super().__init__(dataset=dataset, collate_fn=custom_collate)


def custom_collate(batch: pd.DataFrame) -> pd.DataFrame:
    """Custom collate function for DataLoader, it simply get and return unchanged pandas DF"""
    if isinstance(batch, list):
        return batch[0]

    else:
        return batch



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

template = """from ibm_watson_machine_learning.preprocessing import DataJoinGraph

data_join_graph = DataJoinGraph()

{nodes}

{edges}

data_join_graph.visualize()


from ibm_watson_machine_learning.experiment import AutoAI

experiment = AutoAI(wml_credentials, project_id=project_id)

pipeline_optimizer = experiment.optimizer(
    name={name},
    prediction_type={prediction_type},
    prediction_column={prediction_column},
    scoring={scoring},
    data_join_graph=data_join_graph,
    data_join_only=True
)


# Data join pipeline extracts features from paths:
{paths}
# Finally, selects deduplicated {nb_of_features} features.
# joined_data pandas dataframe can be used to work with joined data.

run_details = pipeline_optimizer.fit(
            training_data_reference=training_data_reference,
            background_mode=False)

joined_data = pipeline_optimizer.get_preprocessed_data_connection().read()
"""



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests
from ibm_watson_machine_learning.utils import SPACES_DETAILS_TYPE, INSTANCE_DETAILS_TYPE, MEMBER_DETAILS_TYPE,DATA_ASSETS_DETAILS_TYPE, STR_TYPE, STR_TYPE_NAME, docstring_parameter, meta_props_str_conv, str_type_conv, get_file_from_cos, modify_details_for_script_and_shiny
from ibm_watson_machine_learning.metanames import ShinyMetaNames
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import WMLClientError, ApiRequestFailure, ForbiddenActionForGitBasedProject
import os

_DEFAULT_LIST_LENGTH = 50


class Shiny(WMLResource):
    """
    Store and manage your shiny assets.

    """
    ConfigurationMetaNames = ShinyMetaNames()
    """MetaNames for Shiny Assets creation."""

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        self._ICP = client.ICP

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_details(self, shiny_uid=None):
        """
            Get shiny asset details. If no shiny_uid is passed, details for all shiny assets will be returned.

            **Parameters**

            .. important::
                #. **shiny_uid**:  Unique id  of shiny asset\n
                   **type**: str\n

            **Output**

            .. important::
                **returns**: Metadata of the stored shiny asset\n
                **return type**: dict\n

            **Example**

             >>> shiny_details = client.shiny.get_details(shiny_uid)

        """
        def get_required_elements(response):
            if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
                return response
            else:
                response = modify_details_for_script_and_shiny(response)
                final_response = {
                    "metadata": response[u'metadata'],
                    "entity": {}
                }

                return final_response

        return self._get_asset_based_resource(shiny_uid, 'shiny_asset', get_required_elements)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def store(self, meta_props, file_path):
        """
                Creates a shiny asset and uploads content to it.

                **Parameters**

                .. important::
                   #. **name**:  Name to be given to the shiny asset\n

                      **type**: str\n

                   #. **file_path**:  Path to the content file to be uploaded\n

                      **type**: str\n

                **Output**

                .. important::

                    **returns**: metadata of the stored shiny asset\n
                    **return type**: dict\n

                **Example**
                 >>> meta_props = {
                 >>>    client.shiny.ConfigurationMetaNames.NAME: "shiny app name"
                 >>> }
                 >>> shiny_details = client.shiny.store(meta_props,file_path="/path/to/file")
        """
        if self._client.project_type == 'local_git_storage':
            raise ForbiddenActionForGitBasedProject(reason="Storing Shiny apps is not supported for git based project.")

        WMLResource._chk_and_block_create_update_for_python36(self)

        Shiny._validate_type(file_path, u'file_path', str, True)

        shiny_meta = self.ConfigurationMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client
        )

        response = self._create_asset(shiny_meta['metadata'], file_path)

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            return response
        else:
            final_response = {
                "metadata": response[u'metadata'],
                "entity": {}
            }

            return final_response

    def _create_asset(self, shiny_meta, file_path):

        ##Step1: Create a shiny asset
        name = shiny_meta[u'name']

        if shiny_meta.get('description') is not None:
            desc = shiny_meta[u'description']
        else:
            desc = ""

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            asset_meta = {
                "metadata": {
                    "name": name,
                    "description": desc,
                    "asset_type": "shiny_asset",
                    "origin_country": "us",
                    "asset_category": "USER"
                }
            }
        else:
            asset_meta = {
                "metadata": {
                    "name": name,
                    "description": desc,
                    "asset_type": "shiny_asset",
                    "origin_country": "us",
                    "asset_category": "USER"
                },
                 "entity": {
                       "shiny_asset": {
                             "ml_version": "4.0.0"
                     }
                 }
            }

        base_shiny_asset = {
            "fields": [
                {
                    "key": "name",
                    "type": "string",
                    "is_array": False,
                    "is_searchable_across_types": False
                },
                {
                    "key": "entity",
                    "type": "object",
                    "is_array": False,
                    "is_searchable_across_types": False
                }
            ],
            "relationships": [],
            "name": "shiny_asset",
            "version": 1
        }

        #Step1  : Create an asset
        print("Creating Shiny asset...")

        if self._client.WSD:
            # For WSD the asset creation is done within _wsd_create_asset function using polyglot
            # Thus using the same for data_assets type


            meta_props = {
                    "name": name
            }
            details = Shiny._wsd_create_asset(self, "shiny_asset", asset_meta, meta_props, file_path, user_archive_file=True)
            return self._get_required_element_from_response(details)
        else:

            creation_response = ""
            if not self._ICP or self._client.ICP_PLATFORM_SPACES:
                creation_response = requests.post(
                        self._client.service_instance._href_definitions.get_data_assets_href(),
                        headers=self._client._get_headers(),
                        params = self._client._params(),
                        json=asset_meta
                )
            else:
                # Until global asset for shiny is created
                asset_type_response = requests.post(
                    self._client.service_instance._href_definitions.get_wsd_asset_type_href() + '?',
                    headers=self._client._get_headers(),
                    json=base_shiny_asset,
                    params=self._client._params()
                )
                if asset_type_response.status_code == 201 or asset_type_response.status_code == 409:
                    creation_response = requests.post(
                        self._client.service_instance._href_definitions.get_data_assets_href(),
                        headers=self._client._get_headers(),
                        json=asset_meta,
                        params=self._client._params()
                    )

            shiny_details = self._handle_response(201, u'creating new asset', creation_response)
            #Step2: Create attachment
            if creation_response.status_code == 201:
                asset_id = shiny_details["metadata"]["asset_id"]
                attachment_meta = {
                        "asset_type": "shiny_asset",
                        "name": "attachment_"+asset_id
                    }

                attachment_response = requests.post(
                    self._client.service_instance._href_definitions.get_attachments_href(asset_id),
                    headers=self._client._get_headers(),
                    params=self._client._params(),
                    json=attachment_meta
                )
                attachment_details = self._handle_response(201, u'creating new attachment', attachment_response)
                if attachment_response.status_code == 201:
                    attachment_id = attachment_details["attachment_id"]
                    attachment_url = attachment_details["url1"]


                    #Step3: Put content to attachment
                    with open(file_path, 'rb') as f:
                        if not self._ICP:
                            put_response = requests.put(
                                attachment_url,
                                data=f.read(),
                                # files = f
                            )
                        else:
                            put_response = requests.put(
                                self._wml_credentials['url'] + attachment_url,
                                files={'file': (name, f, 'application/octet-stream')}
                            )
                    if put_response.status_code == 201 or put_response.status_code == 200:
                        # Step4: Complete attachment
                        complete_response = requests.post(
                            self._client.service_instance._href_definitions.get_attachment_complete_href(asset_id, attachment_id),
                            headers=self._client._get_headers(),
                            params = self._client._params()

                        )

                        if complete_response.status_code == 200:
                            print("SUCCESS")
                            return self._get_required_element_from_response(shiny_details)
                        else:
                            self._delete(asset_id)
                            raise WMLClientError("Failed while creating a shiny asset. Try again.")
                    else:
                        self._delete(asset_id)
                        raise WMLClientError("Failed while creating a shiny asset. Try again.")
                else:
                    print("SUCCESS")
                    return self._get_required_element_from_response(shiny_details)
            else:
                raise WMLClientError("Failed while creating a shiny asset. Try again.")

    def list(self, limit=None):
        """
           List stored shiny assets. If limit is set to None there will be only first 50 records shown.

           **Parameters**

           .. important::
                #. **limit**:  limit number of fetched records\n
                   **type**: int\n

           **Output**

           .. important::
                This method only prints the list of all shiny assets in a table format.\n
                **return type**: None\n

           **Example**

                    >>> client.shiny.list()

        """

        Shiny._validate_type(limit, u'limit', int, False)
        href = self._client.service_instance._href_definitions.get_search_shiny_href()

        data = {
                "query": "*:*"
        }
        if limit is not None:
            data.update({"limit": limit})

        response = requests.post(href, params=self._client._params(), headers=self._client._get_headers(),json=data)
        self._handle_response(200, u'list assets', response)
        shiny_details = self._handle_response(200, u'list assets', response)["results"]
        space_values = [
            (m[u'metadata'][u'name'], m[u'metadata'][u'asset_type'], m["metadata"]["asset_id"]) for
            m in shiny_details]

        self._list(space_values, [u'NAME', u'ASSET_TYPE', u'ASSET_ID'], None, _DEFAULT_LIST_LENGTH)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def download(self, shiny_uid, filename, rev_uid=None):
        """
            Download the content of a shiny asset.

            **Parameters**

            .. important::
                 #. **shiny_uid**:  The Unique Id of the shiny asset to be downloaded\n
                    **type**: str\n

                 #. **filename**:  filename to be used for the downloaded file\n
                    **type**: str\n

                 #. **rev_uid**:  Revision id\n
                    **type**: str\n
            **Output**

                 **returns**: Path to the downloaded shiny asset content\n
                 **return type**: str\n

            **Example**

                  >>> client.shiny.download(shiny_uid,"shiny_asset.zip")

         """
        if rev_uid is not None and self._client.ICP_30 is None and not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u'Not applicable for this release')

        Shiny._validate_type(shiny_uid, u'shiny_uid', STR_TYPE, True)
        Shiny._validate_type(rev_uid, u'rev_uid', int, False)

        params = self._client._params()

        if rev_uid is not None:
            params.update({'revision_id': rev_uid})

        import urllib
        asset_response = requests.get(self._client.service_instance._href_definitions.get_data_asset_href(shiny_uid),
                                      params=params,
                                      headers=self._client._get_headers())
        shiny_details = self._handle_response(200, u'get shiny assets', asset_response)

        if self._WSD:
            attachment_url = shiny_details['attachments'][0]['object_key']
            artifact_content_url = self._client.service_instance._href_definitions.get_wsd_model_attachment_href() + \
                                   urllib.parse.quote('shiny_asset/' + attachment_url, safe='')

            r = requests.get(artifact_content_url, params=self._client._params(), headers=self._client._get_headers(),
                             stream=True)
            if r.status_code != 200:
                raise ApiRequestFailure(u'Failure during {}.'.format("downloading shiny asset"), r)

            downloaded_asset = r.content
            try:
                with open(filename, 'wb') as f:
                    f.write(downloaded_asset)
                print(u'Successfully saved shiny asset content to file: \'{}\''.format(filename))
                return os.getcwd() + "/" + filename
            except IOError as e:
                raise WMLClientError(u'Saving shiny asset with artifact_url: \'{}\'  to local file failed.'.format(filename), e)
        else:
            attachment_id = shiny_details["attachments"][0]["id"]
            response = requests.get(self._client.service_instance._href_definitions.get_attachment_href(shiny_uid,attachment_id), params=params,
                                    headers=self._client._get_headers())
            if response.status_code == 200:
                attachment_signed_url = response.json()["url"]
                if 'connection_id' in shiny_details["attachments"][0]:
                    att_response = requests.get(attachment_signed_url)
                else:
                    if not self._ICP:
                        att_response = requests.get(attachment_signed_url)
                    else:
                        att_response = requests.get(self._wml_credentials["url"]+attachment_signed_url)

                if att_response.status_code != 200:
                    raise ApiRequestFailure(u'Failure during {}.'.format("downloading asset"), att_response)

                downloaded_asset = att_response.content
                try:
                    with open(filename, 'wb') as f:
                        f.write(downloaded_asset)
                    print(u'Successfully saved shiny asset content to file: \'{}\''.format(filename))
                    return os.getcwd() + "/" + filename
                except IOError as e:
                    raise WMLClientError(u'Saving shiny asset with artifact_url to local file: \'{}\' failed.'.format(filename), e)
            else:
                raise WMLClientError("Failed while downloading the shiny asset " + shiny_uid)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_id(shiny_details):
        """
                Get Unique Id of stored shiny asset.

                **Parameters**

                .. important::

                   #. **shiny_details**:  Metadata of the stored shiny asset\n
                      **type**: dict\n
                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: Unique Id of stored shiny asset\n
                    **return type**: str\n

                **Example**

                     >>> shiny_uid = client.shiny.get_id(shiny_details)

        """

        return Shiny.get_uid(shiny_details)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_uid(shiny_details):
        """
                Get Unique Id  of stored shiny asset. This method is deprecated. Use 'get_id(shiny_details)' instead

                **Parameters**

                .. important::

                   #. **shiny_details**:  Metadata of the stored shiny asset\n
                      **type**: dict\n
                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: Unique Id of stored shiny asset\n
                    **return type**: str\n

                **Example**

                     >>> shiny_uid = client.shiny.get_uid(shiny_details)

        """
        Shiny._validate_type(shiny_details, u'shiny_details', object, True)
        Shiny._validate_type_of_details(shiny_details, DATA_ASSETS_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(shiny_details, u'data_assets_details',
                                                           [u'metadata', u'guid'])


    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_href(shiny_details):
        """
            Get url of stored shiny asset.

           **Parameters**

           .. important::
                #. **shiny_details**:  stored shiny asset details\n
                   **type**: dict\n

           **Output**

           .. important::
                **returns**: href of stored shiny asset\n
                **return type**: str\n

           **Example**

             >>> shiny_details = client.shiny.get_details(shiny_uid)
             >>> shiny_href = client.shiny.get_href(shiny_details)
        """
        Shiny._validate_type(shiny_details, u'shiny_details', object, True)
        Shiny._validate_type_of_details(shiny_details, DATA_ASSETS_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(shiny_details, u'shiny_details', [u'metadata', u'href'])

    def update(self, shiny_uid, meta_props=None, file_path=None):
        """
            Update shiny with either metadata or attachment or both.

            :param shiny_uid:  Shiny UID
            :type shiny_uid: str

            :returns: updated metadata of shiny asset
            :rtype: dict

            A way you might use me is:

            >>> script_details = client.script.update(model_uid, meta, content_path)
        """
        WMLResource._chk_and_block_create_update_for_python36(self)
        if self._client.ICP_30 is None and not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u'Not supported in this release')

        Shiny._validate_type(shiny_uid, 'shiny_uid', STR_TYPE, True)

        if meta_props is None and file_path is None:
            raise WMLClientError('Atleast either meta_props or file_path has to be provided')

        updated_details = None
        details = None

        url = self._client.service_instance._href_definitions.get_asset_href(shiny_uid)

        # STEPS
        # STEP 1. Get existing metadata
        # STEP 2. If meta_props provided, we need to patch meta
        #   CAMS has meta and entity patching. 'name' and 'description' get stored in CAMS meta section
        #   a. Construct meta patch string and call /v2/assets/<asset_id> to patch meta
        # STEP 3. If file_path provided, we need to patch the attachment
        #   a. If attachment already exists for the script, delete it
        #   b. POST call to get signed url for upload
        #   c. Upload to the signed url
        #   d. Mark upload complete
        # STEP 4. Get the updated script record and return

        # STEP 1
        response = requests.get(
            url,
            params=self._client._params(),
            headers=self._client._get_headers()
        )

        if response.status_code != 200:
            if response.status_code == 404:
                raise WMLClientError(
                    u'Invalid input. Unable to get the details of shiny_uid provided.')
            else:
                raise ApiRequestFailure(u'Failure during {}.'.format("getting shiny asset to update"), response)

        details = self._handle_response(200, "Get shiny asset details", response)

        attachments_response = None

        # STEP 2a.
        # Patch meta if provided
        if meta_props is not None:
            self._validate_type(meta_props, u'meta_props', dict, True)
            meta_props_str_conv(meta_props)

            meta_patch_payload = []

            # Since we are dealing with direct asset apis, name and description is metadata patch
            if "name" in meta_props or "description" in meta_props:
                props_for_asset_meta_patch = {}

                for key in meta_props:
                    if key == 'name' or key == 'description':
                        props_for_asset_meta_patch.update({key: meta_props[key]})

                meta_patch_payload = self.ConfigurationMetaNames._generate_patch_payload(details,
                                                                                         props_for_asset_meta_patch,
                                                                                         with_validation=True)
            if meta_patch_payload:
                meta_patch_url = self._client.service_instance._href_definitions.get_asset_href(shiny_uid)

                response_patch = requests.patch(meta_patch_url,
                                                json=meta_patch_payload,
                                                params=self._client._params(),
                                                headers=self._client._get_headers())

                updated_details = self._handle_response(200, u'shiny patch', response_patch)

        if file_path is not None:
            if "attachments" in details and details[u'attachments']:
                current_attachment_id = details[u'attachments'][0][u'id']
            else:
                current_attachment_id = None

            #STEP 3
            attachments_response = self._update_attachment_for_assets("shiny_asset",
                                                                      shiny_uid,
                                                                      file_path,
                                                                      current_attachment_id)

        if attachments_response is not None and 'success' not in attachments_response:
            self._update_msg(updated_details)

        # Have to fetch again to reflect updated asset and attachment ids
        url = self._client.service_instance._href_definitions.get_asset_href(shiny_uid)

        response = requests.get(
            url,
            params=self._client._params(),
            headers=self._client._get_headers()
        )

        if response.status_code != 200:
            if response.status_code == 404:
                raise WMLClientError(
                    u'Invalid input. Unable to get the details of shiny_uid provided.')
            else:
                raise ApiRequestFailure(u'Failure during {}.'.format("getting shiny to update"), response)

        # response = self._handle_response(200, "Get shiny details", response)
        #
        # return self._get_required_element_from_response(response)

        response = self._get_required_element_from_response(self._handle_response(200, "Get shiny details", response))

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            return response
        else:
            final_response = {
                "metadata": response[u'metadata'],
                "entity": {}
            }

            return final_response

    def _update_msg(self, updated_details):
        if updated_details is not None:
            print("Could not update the attachment because of server error."
                  " However metadata is updated. Try updating attachment again later")
        else:
            raise WMLClientError('Unable to update attachment because of server error. Try again later')


    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def delete(self, shiny_uid):
        """
            Delete a stored shiny asset.

            **Parameters**

            .. important::
                #. **shiny_uid**:  Unique Id of shiny asset\n
                   **type**: str\n

            **Output**

            .. important::
                **returns**: status ("SUCCESS" or "FAILED")\n
                **return type**: str\n

            **Example**

             >>> client.shiny.delete(shiny_uid)

        """
        Shiny._validate_type(shiny_uid, u'shiny_uid', STR_TYPE, True)
        if (self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES) and \
                self._if_deployment_exist_for_asset(shiny_uid):
            raise WMLClientError(
                u'Cannot delete shiny asset that has existing deployments. Please delete all associated deployments and try again')

        response = requests.delete(self._client.service_instance._href_definitions.get_asset_href(shiny_uid), params=self._client._params(),
                                headers=self._client._get_headers())

        if response.status_code == 200:
            return self._get_required_element_from_response(response.json())
        else:
            return self._handle_response(204, u'delete assets', response)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def create_revision(self, shiny_uid):
        """
           Creates revision for the given Shiny asset. Revisions are immutable once created.
           The metadata and attachment at script_uid is taken and a revision is created out of it

           :param shiny_uid: Shiny asset ID. Mandatory
           :type shiny_uid: {str_type}

           :returns: stored shiny asset revisions metadata
           :rtype: dict

           >>> shiny_revision = client.shiny.create_revision(shiny_uid)
        """
        WMLResource._chk_and_block_create_update_for_python36(self)
        if self._client.ICP_30 is None and not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u'Not supported in this release')

        ##For CP4D, check if either space or project ID is set
        self._client._check_if_either_is_set()

        shiny_uid = str_type_conv(shiny_uid)
        Shiny._validate_type(shiny_uid, u'shiny_uid', STR_TYPE, True)

        print("Creating shiny revision...")
        #
        # return  self._get_required_element_from_response(
        #     self._create_revision_artifact_for_assets(shiny_uid, 'Shiny'))

        response = self._get_required_element_from_response(
            self._create_revision_artifact_for_assets(shiny_uid, 'Shiny'))

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            return response
        else:
            final_response = {
                "metadata": response[u'metadata'],
                "entity": {}
            }

            return final_response

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def list_revisions(self, shiny_uid, limit=None):
        """
           List all revisions for the given shiny asset uid.

           :param shiny_uid: Stored shiny asset ID.
           :type shiny_uid: {str_type}

           :param limit: limit number of fetched records (optional)
           :type limit: int

           :returns: stored shiny revisions details
           :rtype: table

           >>> client.shiny.list_revisions(shiny_uid)
        """

        if self._client.ICP_30 is None and not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u'Not supported in this release')

        ##For CP4D, check if either space or project ID is set
        self._client._check_if_either_is_set()

        shiny_uid = str_type_conv(shiny_uid)
        Shiny._validate_type(shiny_uid, u'shiny_uid', STR_TYPE, True)

        url = self._client.service_instance._href_definitions.get_asset_href(shiny_uid) + "/revisions"
        # /v2/assets/{asset_id}/revisions returns 'results' object
        shiny_resources = self._get_with_or_without_limit(url,
                                                           limit,
                                                           'List Shiny revisions',
                                                           summary=None,
                                                           pre_defined=None)[u'resources']
        shiny_values = [
            (m[u'metadata'][u'asset_id'],
             m[u'metadata'][u'revision_id'],
             m[u'metadata'][u'name'],
             m[u'metadata'][u'commit_info'][u'committed_at']) for m in
            shiny_resources]

        self._list(shiny_values, [u'GUID', u'REVISION_ID', u'NAME', u'REVISION_COMMIT'], limit, _DEFAULT_LIST_LENGTH)

    def get_revision_details(self, shiny_uid=None, rev_uid=None):
        """
           Get metadata of shiny_uid revision.

           :param script_uid: Shiny asset ID. Mandatory
           :type script_uid: {str_type}

           :param rev_uid: Revision ID. If this parameter is not provided, returns latest revision if existing else error
           :type rev_uid: int

           :returns: stored shiny(s) metadata
           :rtype: dict

           A way you might use me is:

           >>> shiny_details = client.shiny.get_revision_details(shiny_uid, rev_uid)
        """
        if self._client.ICP_30 is None and not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError(u'Not supported in this release')

        ##For CP4D, check if either space or project ID is set
        self._client._check_if_either_is_set()

        shiny_uid = str_type_conv(shiny_uid)
        Shiny._validate_type(shiny_uid, u'shiny_uid', STR_TYPE, True)
        Shiny._validate_type(rev_uid, u'rev_uid', int, False)

        if rev_uid is None:
            rev_uid = 'latest'

        url = self._client.service_instance._href_definitions.get_asset_href(shiny_uid)
        # return self._get_required_element_from_response(self._get_with_or_without_limit(url,
        #                                        limit=None,
        #                                        op_name="asset_revision",
        #                                        summary=None,
        #                                        pre_defined=None,
        #                                        revision=rev_uid))
        resources = self._get_with_or_without_limit(url,
                                                    limit=None,
                                                    op_name="asset_revision",
                                                    summary=None,
                                                    pre_defined=None,
                                                    revision=rev_uid)['resources']
        responses = [self._get_required_element_from_response(resource) for resource in resources]

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            return responses
        else:
            final_responses = [{
                "metadata": response[u'metadata'],
                "entity": {}
            } for response in responses]

            return final_responses

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def _delete(self, shiny_uid):
        Shiny._validate_type(shiny_uid, u'shiny_uid', STR_TYPE, True)

        response = requests.delete(self._client.service_instance._href_definitions.get_asset_href(shiny_uid), params=self._client._params(),
                                   headers=self._client._get_headers())


    # @docstring_parameter({'str_type': STR_TYPE_NAME})
    # def get_href(self, shiny_uid):
    #     """
    #        Get metadata of stored space(s). If space UID is not specified, it returns all the spaces metadata.
    #
    #        **Parameters**
    #
    #        .. important::
    #             #. **space_uid**: Space UID (optional)\n
    #                **type**: str\n
    #             #. **limit**:  limit number of fetched records (optional)\n
    #                **type**: int\n
    #
    #        **Output**
    #
    #        .. important::
    #             **returns**: metadata of stored space(s)\n
    #             **return type**: dict
    #             dict (if UID is not None) or {"resources": [dict]} (if UID is None)\n
    #
    #        .. note::
    #             If UID is not specified, all spaces metadata is fetched\n
    #
    #        **Example**
    #
    #         >>> space_details = client.spaces.get_details(space_uid)
    #         >>> space_details = client.spaces.get_details()
    #     """
    #
    #
    #     Assets._validate_type(shiny_uid, u'shiny_uid', STR_TYPE, True)
    #
    #     response = requests.get(self._client.service_instance._href_definitions.get_data_asset_href(shiny_uid), params=self._client._params(),
    #                             headers=self._client._get_headers())
    #
    #     if response.status_code == 200:
    #         return response.json()["href"]
    #     else:
    #         return self._handle_response(200, u'spaces assets', response)

    def _get_required_element_from_response(self, response_data):

        WMLResource._validate_type(response_data, u'shiny', dict)

        revision_id = None

        try:
            if self._client.default_space_id is not None:
                metadata = {'space_id': response_data['metadata']['space_id'],
                            'guid': response_data['metadata']['asset_id'],
                            'href': response_data['href'],
                            'name': response_data[u'metadata'][u'name'],
                            'asset_type': response_data['metadata']['asset_type'],
                            'created_at': response_data['metadata']['created_at'],
                            'last_updated_at': response_data['metadata']['usage']['last_updated_at']
                            }
                if self._client.ICP_30 is not None or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                    if "revision_id" in response_data[u'metadata']:
                        revision_id = response_data[u'metadata'][u'revision_id']
                        metadata.update({'revision_id': response_data[u'metadata'][u'revision_id']})

                    if "attachments" in response_data and response_data[u'attachments']:
                        metadata.update({'attachment_id': response_data[u'attachments'][0][u'id']})

                    if "commit_info" in response_data[u'metadata'] and revision_id is not None:
                        metadata.update(
                            {'revision_commit_date': response_data[u'metadata'][u'commit_info']['committed_at']})

                new_el = {'metadata': metadata,
                          'entity': response_data['entity']
                          }

            elif self._client.default_project_id is not None:
                if self._client.WSD:

                    href = self._client.service_instance._href_definitions.get_base_asset_href(response_data['metadata']['asset_id']) + "?" + "project_id=" + response_data['metadata']['project_id']

                    metadata = {'project_id': response_data['metadata']['project_id'],
                                'guid': response_data['metadata']['asset_id'],
                                'href': href,
                                'name': response_data[u'metadata'][u'name'],
                                'asset_type': response_data['metadata']['asset_type'],
                                'created_at': response_data['metadata']['created_at']
                    }
                    if self._client.WSD_20 is not None:
                        if "revision_id" in response_data[u'metadata']:
                            revision_id = response_data[u'metadata'][u'revision_id']
                            metadata.update({'revision_id': response_data[u'metadata'][u'revision_id']})

                        if "attachments" in response_data and response_data[u'attachments']:
                            metadata.update({'attachment_id': response_data[u'attachments'][0][u'id']})

                        if "commit_info" in response_data[u'metadata'] and revision_id is not None:
                            metadata.update(
                                {'revision_commit_date': response_data[u'metadata'][u'commit_info']['committed_at']})

                    new_el = {'metadata': metadata,
                              'entity': response_data['entity']
                              }
                    if 'usage' in response_data['metadata']:
                        new_el['metadata'].update(
                            {'last_updated_at': response_data['metadata']['usage']['last_updated_at']})
                    else:
                        new_el['metadata'].update(
                            {'last_updated_at': response_data['metadata']['last_updated_at']})

                else:
                    metadata = {'project_id': response_data['metadata']['project_id'],
                                'guid': response_data['metadata']['asset_id'],
                                'href': response_data['href'],
                                'name': response_data[u'metadata'][u'name'],
                                'asset_type': response_data['metadata']['asset_type'],
                                'created_at': response_data['metadata']['created_at'],
                                'last_updated_at': response_data['metadata']['usage']['last_updated_at']
                                }

                    if self._client.ICP_30 is not None or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                        if "revision_id" in response_data[u'metadata']:
                            revision_id = response_data[u'metadata'][u'revision_id']
                            metadata.update({'revision_id': response_data[u'metadata'][u'revision_id']})

                        if "attachments" in response_data and response_data[u'attachments']:
                            metadata.update({'attachment_id': response_data[u'attachments'][0][u'id']})

                        if "commit_info" in response_data[u'metadata'] and revision_id is not None:
                            metadata.update(
                                {'revision_commit_date': response_data[u'metadata'][u'commit_info']['committed_at']})

                    new_el = {'metadata': metadata,
                              'entity': response_data['entity']
                              }
            if 'description' in response_data['metadata']:
                new_el[u'metadata'].update({'description': response_data[u'metadata'][u'description']})

            if (self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES) and 'href' in response_data['metadata']:
                href_without_host = response_data['metadata']['href'].split('.com')[-1]
                new_el[u'metadata'].update({'href': href_without_host})
            return new_el
        except Exception as e:
            raise WMLClientError("Failed to read Response from down-stream service: " + response_data.text)



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests
from ibm_watson_machine_learning.utils import STR_TYPE, STR_TYPE_NAME, docstring_parameter, meta_props_str_conv
from ibm_watson_machine_learning.metanames import ExportMetaNames
import os
import json
from ibm_watson_machine_learning.wml_client_error import WMLClientError, UnexpectedType, ApiRequestFailure
from ibm_watson_machine_learning.wml_resource import WMLResource
_DEFAULT_LIST_LENGTH = 50

class Export(WMLResource):
    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)

        self._client = client
        self.ConfigurationMetaNames = ExportMetaNames()

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def start(self, meta_props, space_id=None, project_id=None):
        """
            Start the export. Either space_id or project_id has to be provided and is mandatory.
            ALL_ASSETS is by default False. No need to provide explicitly unless it has to be set to True
            Either ALL_ASSETS or ASSET_TYPES or ASSET_IDS has to be given in the meta_props. Only one of these can be
            provided

            In the meta_props:

            ALL_ASSETS is a boolean. When set to True, it exports all assets in the given space
            ASSET_IDS is an array containing the list of assets ids to be exported
            ASSET_TYPES is for providing the asset types to be exported. All assets of that asset type will be exported
                        Eg: wml_model, wml_model_definition, wml_pipeline, wml_function, wml_experiment,
                        software_specification, hardware_specification, package_extension, script

           **Parameters**

           .. important::
                #. **meta_props**:  meta data. To see available meta names use **client.export_assets.ConfigurationMetaNames.get()**\n
                   **type**: dict\n
                #. **space_id**:  Space identifier**\n
                   **type**: str\n
                #. **project_id**:  Project identifier**\n
                   **type**: str\n

           **Output**

           .. important::
                **returns**: Response json\n
                **return type**: dict\n

           **Example**

                >>> metadata = {
                >>>    client.export_assets.ConfigurationMetaNames.NAME: "export_model",
                >>>    client.export_assets.ConfigurationMetaNames.ASSET_IDS: ["13a53931-a8c0-4c2f-8319-c793155e7517", "13a53931-a8c0-4c2f-8319-c793155e7518"]
                >>> }
                >>> details = client.export_assets.start(meta_props=metadata, space_id="98a53931-a8c0-4c2f-8319-c793155e4598")

                >>> metadata = {
                >>>    client.export_assets.ConfigurationMetaNames.NAME: "export_model",
                >>>    client.export_assets.ConfigurationMetaNames.ASSET_TYPES: ["wml_model"],
                >>> }
                >>> details = client.export_assets.start(meta_props=metadata, space_id="98a53931-a8c0-4c2f-8319-c793155e4598")

                >>> metadata = {
                >>>    client.export_assets.ConfigurationMetaNames.NAME: "export_model",
                >>>    client.export_assets.ConfigurationMetaNames.ALL_ASSETS: True,
                >>> }
            """
        WMLResource._chk_and_block_create_update_for_python36(self)

        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("Not supported in this release")

        if space_id is None and project_id is None:
            raise WMLClientError("Either 'space_id' or 'project_id' has to be provided")

        if space_id is not None and project_id is not None:
            raise WMLClientError("Either 'space_id' or 'project_id' can be provided, not both")

        Export._validate_type(meta_props, u'meta_props', dict, True)
        self._validate_input_meta(meta_props)

        meta = self.ConfigurationMetaNames._generate_resource_metadata(
            meta_props,
            with_validation=True,
            client=self._client)

        start_meta = {}
        assets = {}

        start_meta[u'name'] = meta[u'name']
        if 'description' in meta:
            start_meta[u'description'] = meta[u'description']

        if "all_assets" not in meta:
            assets.update({"all_assets": False})
        else:
            assets.update({"all_assets": meta[u'all_assets']})

        if "asset_types" in meta:
            assets.update({"asset_types": meta[u'asset_types']})

        if "asset_ids" in meta:
            assets.update(({"asset_ids": meta[u'asset_ids']}))

        start_meta[u'assets'] = assets

        href = self._client.service_instance._href_definitions.exports_href()

        params = {}

        if space_id is not None:
            params.update({"space_id": space_id})
        else:
            params.update({"project_id": project_id})

        creation_response = requests.post(href,
                                          params=params,
                                          headers=self._client._get_headers(),
                                          json=start_meta)

        details = self._handle_response(expected_status_code=202,
                                        operationName=u'export start',
                                        response=creation_response)

        export_id = details[u'metadata']['id']

        print("export job with id {} has started. Monitor status using client.export_assets.get_details api. "
              "Check 'help(client.export_assets.get_details)' for details on the api usage".format(export_id))

        return details

    def _validate_input_meta(self, meta_props):
        if 'name' not in meta_props:
            raise WMLClientError("Its mandatory to provide 'NAME' in meta_props. Example: "
                                 "client.export_assets.ConfigurationMetaNames.NAME: 'name'")

        if 'all_assets' not in meta_props and 'asset_ids' not in meta_props and 'asset_types' not in meta_props:
            raise WMLClientError("Its mandatory to provide either 'ALL_ASSETS' or 'ASSET_IDS' or 'ASSET_TYPES' " 
                                 "in meta_props. Example: client.export_assets.ConfigurationMetaNames.ALL_ASSETS: True")

        count = 0

        if 'all_assets' in meta_props:
            count = count + 1
        if 'asset_ids' in meta_props:
            count = count + 1
        if 'asset_types' in meta_props:
            count = count + 1

        if count > 1:
            raise WMLClientError("Only one of 'ALL_ASSETS' or 'ASSET_IDS' or 'ASSET_TYPES' can be provided")

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def cancel(self, export_id, space_id=None, project_id=None):
        """
            Cancel an export job. 'export_id' and 'space_id'( or 'project_id' has to be provided )
            Note: To delete a export_id job, use delete() api

            **Parameters**

            .. important::
                #. **export_id**:  Export job identifier\n
                   **type**: str\n
                #. **space_id**:  Space identifier\n
                   **type**: str\n
                #. **project_id**:  Project identifier\n
                   **type**: str\n

            **Output**

            .. important::
                **returns**: status ("SUCCESS" or "FAILED")\n
                **return type**: str\n

            **Example**

             >>> client.export_assets.cancel(export_id='6213cf1-252f-424b-b52d-5cdd9814956c',
             >>>                      space_id='3421cf1-252f-424b-b52d-5cdd981495fe')
        """
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("Not supported in this release")

        Export._validate_type(export_id, u'export_id', STR_TYPE, True)

        if space_id is None and project_id is None:
            raise WMLClientError("Its mandatory to provide space_id or project_id")

        if space_id is not None and project_id is not None:
            raise WMLClientError("Either 'space_id' or 'project_id' can be provided, not both")

        href = self._client.service_instance._href_definitions.export_href(export_id)

        params = {}

        if space_id is not None:
            params.update({"space_id": space_id})
        else:
            params.update({"project_id": project_id})

        cancel_response = requests.delete(href,
                                          params=params,
                                          headers=self._client._get_headers())

        details = self._handle_response(expected_status_code=204,
                                        operationName=u'cancel export',
                                        response=cancel_response)

        if "SUCCESS" == details:
            print("Export job cancelled")

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def delete(self, export_id, space_id=None, project_id=None):
        """
            Deletes the given export_id job. 'space_id' or 'project_id' has to be provided

            **Parameters**

            .. important::
                #. **export_id**:  Export job identifier\n
                   **type**: str\n
                #. **space_id**:  Space identifier**\n
                   **type**: str\n
                #. **project_id**:  Project identifier**\n
                   **type**: str\n

            **Output**

            .. important::
                **returns**: status ("SUCCESS" or "FAILED")\n
                **return type**: str\n

            **Example**

             >>> client.export_assets.delete(export_id='6213cf1-252f-424b-b52d-5cdd9814956c',
             >>>                      space_id= '98a53931-a8c0-4c2f-8319-c793155e4598')
        """
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("Not supported in this release")

        if space_id is None and project_id is None:
            raise WMLClientError("Its mandatory to provide space_id or project_id")

        if space_id is not None and project_id is not None:
            raise WMLClientError("Either 'space_id' or 'project_id' can be provided, not both")

        Export._validate_type(export_id, u'export_id', STR_TYPE, True)

        href = self._client.service_instance._href_definitions.export_href(export_id)

        params = {"hard_delete": True}

        if space_id is not None:
            params.update({'space_id': space_id})
        else:
            params.update({'project_id': project_id})

        delete_response = requests.delete(href,
                                          params=params,
                                          headers=self._client._get_headers())

        details = self._handle_response(expected_status_code=204,
                                        operationName=u'delete export job',
                                        response=delete_response)

        if "SUCCESS" == details:
            print("Export job deleted")

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_details(self, export_id=None, space_id=None, project_id=None, limit=None, asynchronous=False, get_all=False):
        """
            Get metadata of the given export job. if no export_id is specified all
            exports metadata is returned.

           **Parameters**

           .. important::
                #. **export_id**: export job identifier (optional)\n
                   **type**: str\n
                #. **space_id**:  Space identifier**\n
                   **type**: str\n
                #. **project_id**:  Project identifier**\n
                   **type**: str\n
                #. **limit**:  limit number of fetched records (optional)\n
                   **type**: int\n
                #. **asynchronous**:  if True, it will work as a generator (optional)\n
                   **type**: bool\n
                #. **get_all**:  if True, it will get all entries in 'limited' chunks (optional)\n
                   **type**: bool\n

           **Output**

           .. important::
                **returns**: export(s) metadata\n
                **return type**: dict\n
                dict (if export_id is not None) or {"resources": [dict]} (if export_id is None)\n

           .. note::
                If export_id is not specified, all export(s) metadata is fetched\n

           **Example**

             >>> details = client.export_assets.get_details(export_id, space_id= '98a53931-a8c0-4c2f-8319-c793155e4598')
             >>> details = client.export_assets.get_details()
             >>> details = client.export_assets.get_details(limit=100)
             >>> details = client.export_assets.get_details(limit=100, get_all=True)
             >>> details = []
             >>> for entry in client.export_assets.get_details(limit=100, asynchronous=True, get_all=True):
             >>>    details.extend(entry)
         """
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("Not supported in this release")

        if space_id is None and project_id is None:
            raise WMLClientError("Its mandatory to provide space_id or project_id")

        if space_id is not None and project_id is not None:
            raise WMLClientError("Either 'space_id' or 'project_id' can be provided, not both")
        
        Export._validate_type(export_id, u'export_id', STR_TYPE, False)
        Export._validate_type(limit, u'limit', int, False)

        href = self._client.service_instance._href_definitions.exports_href()

        params = {}

        if space_id is not None:
            params.update({"space_id": space_id})
        else:
            params.update({"project_id": project_id})

        if export_id is None:
            return self._get_artifact_details(href, export_id, limit, 'export job', query_params=params,
                                              _async=asynchronous, _all=get_all)

        else:
            return self._get_artifact_details(href, export_id, limit, 'export job', query_params=params)
    
    def list(self, space_id=None, project_id=None, limit=None):
        """
            List export jobs. If limit is set to None there will be only first 50 records shown.

           **Parameters**

           .. important::
                #. **limit**:  limit number of fetched records\n
                   **type**: int\n
                #. **space_id**:  Space identifier**\n
                   **type**: str\n
                #. **project_id**:  Project identifier**\n
                   **type**: str\n

           **Output**

           .. important::
                This method only prints the list of all export jobs in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.export_assets.list()
        """
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("Not supported in this release")

        if space_id is None and project_id is None:
            raise WMLClientError("Its mandatory to provide space_id or project_id")

        if space_id is not None and project_id is not None:
            raise WMLClientError("Either 'space_id' or 'project_id' can be provided, not both")

        if space_id is not None:
            resources = self.get_details(space_id=space_id)[u'resources']
        else:
            resources = self.get_details(project_id=project_id)[u'resources']


        values = [(m[u'metadata'][u'id'],
                   m[u'metadata'][u'name'],
                   m[u'metadata'][u'created_at'],
                   m[u'entity'][u'status'][u'state']) for m in resources]

        self._list(values, [u'ID', u'NAME', u'CREATED', u'STATUS'], limit, _DEFAULT_LIST_LENGTH)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_id(export_details):
        """
                Get ID of export job from export details

                **Parameters**

                .. important::

                   #. **export_details**:  Metadata of the export job\n
                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: ID of the export job\n
                    **return type**: str\n

                **Example**

                 >>> id = client.export_assets.get_id(export_details)
        """
        Export._validate_type(export_details, u'export_details', object, True)

        return WMLResource._get_required_element_from_dict(export_details,
                                                           u'export_details',
                                                           [u'metadata', u'id'])

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_exported_content(self, export_id, space_id=None, project_id=None, file_path=None):
        """
            Get the exported content as a zip file

            **Parameters**

            .. important::
                #. **export_id**:  export job identifier\n
                   **type**: str\n
                #. **space_id**:  Space identifier**\n
                   **type**: str\n
                #. **project_id**:  Project identifier**\n
                   **type**: str\n
                #. **file_path**:  name of local file to create. This should be absolute path of the file and
                                  the file shouldn't exist**\n
                   **type**: str\n


            **Output**

            .. important::

               **returns**: Path to the downloaded function content\n
               **return type**: str\n

            **Example**

             >>> client.exports.get_exported_content(export_id,
             >>>                                     space_id= '98a53931-a8c0-4c2f-8319-c793155e4598'
             >>>                                     file_path='/home/user/my_exported_content.zip')
        """
        if not self._client.CLOUD_PLATFORM_SPACES and not self._client.ICP_PLATFORM_SPACES:
            raise WMLClientError("Not supported in this release")

        if space_id is None and project_id is None:
            raise WMLClientError("Its mandatory to provide space_id or project_id")

        if space_id is not None and project_id is not None:
            raise WMLClientError("Either 'space_id' or 'project_id' can be provided, not both")

        params = {}

        if space_id is not None:
            params.update({"space_id": space_id})
        else:
            params.update({"project_id": project_id})

        if os.path.isfile(file_path):
            raise WMLClientError(u'File with name: \'{}\' already exists.'.format(file_path))

        Export._validate_type(file_path, u'file_path', STR_TYPE, True)

        href = self._client.service_instance._href_definitions.export_content_href(export_id)

        try:
            response = requests.get(href,
                                    params=params,
                                    headers=self._client._get_headers(),
                                    stream=True)

            if response.status_code != 200:
                raise ApiRequestFailure(u'Failure during {}.'.format("downloading export content"), response)

            downloaded_exported_content = response.content
            self._logger.info(u'Successfully downloaded artifact with artifact_url: {}'.format(href))
        except WMLClientError as e:
            raise e
        except Exception as e:
            raise WMLClientError(u'Downloading export content with artifact_url: \'{}\' failed.'.format(href), e)

        try:
            with open(file_path, 'wb') as f:
                f.write(downloaded_exported_content)
            print(u'Successfully saved export content to file: \'{}\''.format(file_path))
            return file_path
        except IOError as e:
            raise WMLClientError(u'Downloading export content with artifact_url: \'{}\' failed.'.format(href), e)




#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import print_function
import ibm_watson_machine_learning._wrappers.requests as requests

from ibm_watson_machine_learning.messages.messages import Messages
from ibm_watson_machine_learning.utils import SPACES_DETAILS_TYPE, INSTANCE_DETAILS_TYPE, MEMBER_DETAILS_TYPE,DATA_ASSETS_DETAILS_TYPE, STR_TYPE, STR_TYPE_NAME, docstring_parameter, meta_props_str_conv, str_type_conv, get_file_from_cos
from ibm_watson_machine_learning.metanames import AssetsMetaNames
from ibm_watson_machine_learning.wml_resource import WMLResource
from ibm_watson_machine_learning.wml_client_error import WMLClientError, ApiRequestFailure
import os

_DEFAULT_LIST_LENGTH = 50


class Assets(WMLResource):
    """
    Store and manage your data assets.
    """
    ConfigurationMetaNames = AssetsMetaNames()
    """MetaNames for Data Assets creation."""

    def __init__(self, client):
        WMLResource.__init__(self, __name__, client)
        self._ICP = client.ICP

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_details(self, asset_uid=None):
        """
            Get data asset details. If no asset_uid is passed, details for all assets will be returned.

            **Parameters**

            .. important::
                **returns**: Unique id  of asset\n
                **return type**: str

            **Output**

            .. important::
                #. **asset_details**:  Metadata of the stored data asset\n
                   **type**: dict\n

            **Example**

             >>> asset_details = client.data_assets.get_details(asset_uid)
        """
        return self._get_asset_based_resource(asset_uid, 'data_asset', self._get_required_element_from_response)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def create(self, name, file_path):
        """
                Creates a data asset and uploads content to it.

                **Parameters**

                .. important::
                   #. **name**:  Name to be given to the data asset\n

                      **type**: str\n

                   #. **file_path**:  Path to the content file to be uploaded\n

                      **type**: str\n

                **Output**

                .. important::

                    **returns**: metadata of the stored data asset\n
                    **return type**: dict\n

                **Example**

                 >>> asset_details = client.data_assets.create(name="sample_asset",file_path="/path/to/file")
        """
        # quick support for COS credentials instead of local path
        # TODO add error handling and cleaning (remove the file)
        WMLResource._chk_and_block_create_update_for_python36(self)
        Assets._validate_type(name, u'name', str, True)
        Assets._validate_type(file_path, u'file_path', str, True)
        return self._create_asset(name, file_path)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def store(self, meta_props):
        """
                Creates a data asset and uploads content to it.

                **Parameters**

                .. important::
                   #. **meta_props**:  meta data of the space configuration. To see available meta names use:\n
                                    >>> client.data_assets.ConfigurationMetaNames.get()

                      **type**: dict\n

                **Example**
                    Example for data asset creation for files :

                            >>> metadata = {
                            >>>  client.data_assets.ConfigurationMetaNames.NAME: 'my data assets',
                            >>>  client.data_assets.ConfigurationMetaNames.DESCRIPTION: 'sample description',
                            >>>  client.data_assets.ConfigurationMetaNames.DATA_CONTENT_NAME: 'sample.csv'
                            >>> }
                            >>> asset_details = client.data_assets.store(meta_props=metadata)

                    Example of data asset creation using connection:

                            >>> metadata = {
                            >>>  client.data_assets.ConfigurationMetaNames.NAME: 'my data assets',
                            >>>  client.data_assets.ConfigurationMetaNames.DESCRIPTION: 'sample description',
                            >>>  client.data_assets.ConfigurationMetaNames.CONNECTION_ID: '39eaa1ee-9aa4-4651-b8fe-95d3ddae',
                            >>>  client.data_assets.ConfigurationMetaNames.DATA_CONTENT_NAME: 't1/sample.csv'
                            >>> }
                            >>> asset_details = client.data_assets.store(meta_props=metadata)

                    Example for data asset creation with database sources type connection:

                            >>> metadata = {
                            >>>  client.data_assets.ConfigurationMetaNames.NAME: 'my data assets',
                            >>>  client.data_assets.ConfigurationMetaNames.DESCRIPTION: 'sample description',
                            >>>  client.data_assets.ConfigurationMetaNames.CONNECTION_ID: '23eaf1ee-96a4-4651-b8fe-95d3dadfe',
                            >>>  client.data_assets.ConfigurationMetaNames.DATA_CONTENT_NAME: 't1'
                            >>> }
                            >>> asset_details = client.data_assets.store(meta_props=metadata)
        """
        ##For CP4D, check if either spce or project ID is set
        WMLResource._chk_and_block_create_update_for_python36(self)
        self._client._check_if_either_is_set()

        # quick support for COS credentials instead of local path
        # TODO add error handling and cleaning (remove the file)
        Assets._validate_type(meta_props, u'meta_props', dict, True)

        name = meta_props[self.ConfigurationMetaNames.NAME]
        file_path = meta_props[self.ConfigurationMetaNames.DATA_CONTENT_NAME]
        connection_id = None
        description = ""
        if self.ConfigurationMetaNames.CONNECTION_ID in meta_props:
            connection_id = meta_props[self.ConfigurationMetaNames.CONNECTION_ID]

        if self.ConfigurationMetaNames.DESCRIPTION in meta_props:
            description = meta_props[self.ConfigurationMetaNames.DESCRIPTION]

        return self._create_asset(name, file_path, connection_id=connection_id, description=description)

    def _create_asset(self, name, file_path, connection_id=None, description=None):
        ##Step1: Create a data asset
        desc = description
        if desc is None:
            desc = ""
        try:
            import mimetypes
        except Exception as e:
            raise WMLClientError(Messages.get_message(message_id="module_mimetypes_not_found"), e)
        mime_type = mimetypes.MimeTypes().guess_type(file_path)[0]
        if mime_type is None:
            mime_type = "application/octet-stream"

        asset_meta = {
            "metadata": {
                "name": name,
                "description": desc,
                "asset_type": "data_asset",
                "origin_country": "us",
                "asset_category": "USER"
            },
            "entity": {
                "data_asset": {
                    "mime_type": mime_type
                }
            }
        }
        if connection_id is not None:
            asset_meta["metadata"].update({"tags": ["connected-data"]})

        #Step1  : Create an asset
        print(Messages.get_message(message_id="creating_data_asset"))

        if self._client.WSD:
            # For WSD the asset creation is done within _wsd_create_asset function using polyglot
            # Thus using the same for data_assets type
            input_payload = {
                    "name": name,
                    "mime_type": mime_type
                    }

            meta_props = {
                    "name": name
            }
            details = Assets._wsd_create_asset(self, "data_asset", input_payload, meta_props, file_path)
            return self._get_required_element_from_response(details)
        else:

            creation_response = requests.post(
                    self._client.service_instance._href_definitions.get_data_assets_href(),
                    headers=self._client._get_headers(),
                    params = self._client._params(),
                    json=asset_meta
            )


            asset_details = self._handle_response(201, u'creating new asset', creation_response)
            #Step2: Create attachment
            if creation_response.status_code == 201:
                asset_id = asset_details["metadata"]["asset_id"]
                attachment_name = file_path.split("/")[-1]
                attachment_meta = {
                        "asset_type": "data_asset",
                        "name": attachment_name,
                        "mime": mime_type
                    }
                if connection_id is not None:
                    attachment_meta.update({"connection_id": connection_id,
                                            "connection_path": file_path,
                                            "is_remote": True})

                attachment_response = requests.post(
                    self._client.service_instance._href_definitions.get_attachments_href(asset_id),
                    headers=self._client._get_headers(),
                    params=self._client._params(),
                    json=attachment_meta
                )
                attachment_details = self._handle_response(201, u'creating new attachment', attachment_response)
                if attachment_response.status_code == 201:
                    if connection_id is None:
                        attachment_id = attachment_details["attachment_id"]
                        attachment_url = attachment_details["url1"]
                        #Step3: Put content to attachment
                        with open(file_path, 'rb') as _file:
                            if not self._ICP:
                                    put_response = requests.put(
                                        attachment_url,
                                        data=_file
                                    )
                            else:
                                    put_response = requests.put(
                                        self._wml_credentials['url'] + attachment_url,
                                        files={'file': (name, _file, 'file')}
                                    )
                        if put_response.status_code == 201 or put_response.status_code == 200:
                            # Step4: Complete attachment

                            complete_response = requests.post(
                                self._client.service_instance._href_definitions.get_attachment_complete_href(asset_id, attachment_id),
                                headers=self._client._get_headers(),
                                params = self._client._params()

                            )

                            if complete_response.status_code == 200:
                                print(Messages.get_message(message_id="success"))
                                return self._get_required_element_from_response(asset_details)
                            else:
                                self._delete(asset_id)
                                raise WMLClientError(Messages.get_message(message_id="failed_while_creating_a_data_asset"))
                        else:
                            self._delete(asset_id)
                            raise WMLClientError(Messages.get_message(message_id="failed_while_creating_a_data_asset"))
                    else:
                        print(Messages.get_message(message_id="success"))
                        return self._get_required_element_from_response(asset_details)
                else:
                    self._delete(asset_id)
                    raise WMLClientError(Messages.get_message(message_id="failed_while_creating_a_data_asset"))
            else:
                raise WMLClientError(Messages.get_message(message_id="failed_while_creating_a_data_asset"))

    def list(self, limit=None):
        """
           List stored data assets. If limit is set to None there will be only first 50 records shown.

           **Parameters**

           .. important::
                #. **limit**:  limit number of fetched records\n
                   **type**: int\n

           **Output**

           .. important::
                This method only prints the list of all data assets in a table format.\n
                **return type**: None\n

           **Example**

            >>> client.data_assets.list()
        """

        Assets._validate_type(limit, u'limit', int, False)
        href = self._client.service_instance._href_definitions.get_search_asset_href()

        data = {
                "query": "*:*"
        }
        if limit is not None:
            data.update({"limit": limit})

        response = requests.post(href, params=self._client._params(), headers=self._client._get_headers(),json=data)
        self._handle_response(200, u'list assets', response)
        asset_details = self._handle_response(200, u'list assets', response)["results"]
        space_values = [
            (m[u'metadata'][u'name'], m[u'metadata'][u'asset_type'], m[u'metadata'][u'size'], m["metadata"]["asset_id"]) for
            m in asset_details]

        self._list(space_values, [u'NAME', u'ASSET_TYPE', u'SIZE', u'ASSET_ID'], limit, _DEFAULT_LIST_LENGTH)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def download(self, asset_uid, filename):
        """
            Download and store the content of a data asset.

            **Parameters**

            .. important::

                 #. **asset_uid**:  The Unique Id of the data asset to be downloaded\n
                    **type**: str\n

                 #. **filename**:  filename to be used for the downloaded file\n
                    **type**: str\n

            **Output**

            .. important::

                 **returns**: Normalized path to the downloaded asset content\n
                 **return type**: str\n

            **Example**

             >>> client.data_assets.download(asset_uid,"sample_asset.csv")
         """
        content = self.get_content(asset_uid)
        try:
            with open(filename, 'wb') as f:
                f.write(content)
            print(Messages.get_message(filename, message_id="successfully_saved_data_asset_content_to_file"))
            return os.path.abspath(filename)
        except IOError as e:
            raise WMLClientError(
                Messages.get_message(filename, message_id="saving_data_asset_to_local_file_failed"), e)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_content(self, asset_uid):
        """
            Download the content of a data asset.

            **Parameters**

            .. important::

                 #. **asset_uid**:  The Unique Id of the data asset to be downloaded\n
                    **type**: str\n

            **Output**

            .. important::

                 **returns**: the asset content\n
                 **return type**: binary\n

            **Example**

             >>> content = client.data_assets.get_content(asset_uid).decode('ascii')
        """
        Assets._validate_type(asset_uid, u'asset_uid', STR_TYPE, True)

        import urllib
        asset_response = requests.get(self._client.service_instance._href_definitions.get_data_asset_href(asset_uid),
                                      params=self._client._params(),
                                      headers=self._client._get_headers())
        asset_details = self._handle_response(200, u'get assets', asset_response)

        if self._WSD:
            attachment_url = asset_details['attachments'][0]['object_key']
            artifact_content_url = self._client.service_instance._href_definitions.get_wsd_model_attachment_href() + \
                                   urllib.parse.quote('data_asset/' + attachment_url, safe='')

            r = requests.get(artifact_content_url, params=self._client._params(), headers=self._client._get_headers(),
                             stream=True)
            if r.status_code != 200:
                raise ApiRequestFailure(Messages.get_message(message_id="failure_during_downloading_data_asset"), r)

            return r.content
        else:
            attachment_id = asset_details["attachments"][0]["id"]
            response = requests.get(self._client.service_instance._href_definitions.get_attachment_href(asset_uid,attachment_id),
                                    params=self._client._params(),
                                    headers=self._client._get_headers())
            if response.status_code == 200:
                if 'connection_id' in asset_details["attachments"][0] and \
                        asset_details["attachments"][0]['connection_id'] is not None:

                    conn_details = self._client.connections.get_details(asset_details["attachments"][0]['connection_id'])
                    attachment_data_source_type = conn_details['entity'].get('datasource_type')
                    cos_conn_data_source_id = self._client.connections.get_datasource_type_uid_by_name('cloudobjectstorage')
                    if attachment_data_source_type == cos_conn_data_source_id:
                        attachment_signed_url = response.json()["url"]
                        att_response = requests.get(attachment_signed_url)
                    else:
                        raise WMLClientError(Messages.get_message(message_id="download_api_not_supported_for_this_connection_type"))
                else:
                    attachment_signed_url = response.json()["url"]
                    if not self._ICP and not self._client.WSD:
                        if self._client.CLOUD_PLATFORM_SPACES:
                            att_response = requests.get(attachment_signed_url)
                        else:
                            att_response = requests.get(self._wml_credentials["url"]+attachment_signed_url)
                    else:
                        att_response = requests.get(self._wml_credentials["url"]+attachment_signed_url)
                if att_response.status_code != 200:
                    raise ApiRequestFailure(Messages.get_message(message_id="failure_during_downloading_data_asset"), att_response)

                return att_response.content
            else:
                raise WMLClientError(Messages.get_message(message_id="failure_during_downloading_data_asset"))

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_uid(asset_details):
        """
                Get Unique Id  of stored data asset. Deprecated!! Use get_id(details) instead

                **Parameters**

                .. important::

                   #. **asset_details**:  Metadata of the stored data asset\n
                      **type**: dict\n
                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: Unique Id of stored asset\n
                    **return type**: str\n

                **Example**


                 >>> asset_uid = client.data_assets.get_uid(asset_details)
        """
        Assets._validate_type(asset_details, u'asset_details', object, True)
        Assets._validate_type_of_details(asset_details, DATA_ASSETS_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(asset_details, u'data_assets_details',
                                                           [u'metadata', u'guid'])


    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_id(asset_details):
        """
                Get Unique Id of stored script asset.

                **Parameters**

                .. important::

                   #. **asset_details**:  Metadata of the stored script asset\n
                      **type**: dict\n
                      **type**: dict\n

                **Output**

                .. important::

                    **returns**: Unique Id of stored data asset\n
                    **return type**: str\n

                **Example**

                     >>> asset_id = client.data_assets.get_id(asset_details)

        """

        return Assets.get_uid(asset_details)

    @staticmethod
    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def get_href(asset_details):
        """
            Get url of stored data asset.

           **Parameters**

           .. important::
                #. **asset_details**:  stored data asset details\n
                   **type**: dict\n

           **Output**

           .. important::
                **returns**: href of stored data asset\n
                **return type**: str\n

           **Example**

             >>> asset_details = client.data_assets.get_details(asset_uid)
             >>> asset_href = client.data_assets.get_href(asset_details)
        """
        Assets._validate_type(asset_details, u'asset_details', object, True)
        Assets._validate_type_of_details(asset_details, DATA_ASSETS_DETAILS_TYPE)

        return WMLResource._get_required_element_from_dict(asset_details, u'asset_details', [u'metadata', u'href'])


    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def delete(self, asset_uid):
        """
            Delete a stored data asset.

            **Parameters**

            .. important::
                #. **asset_uid**:  Unique Id of data asset\n
                   **type**: str\n

            **Output**

            .. important::
                **returns**: status ("SUCCESS" or "FAILED")\n
                **return type**: str\n

            **Example**

             >>> client.data_assets.delete(asset_uid)
        """
        Assets._validate_type(asset_uid, u'asset_uid', STR_TYPE, True)

        response = requests.delete(self._client.service_instance._href_definitions.get_asset_href(asset_uid), params=self._client._params(),
                                headers=self._client._get_headers())
        if response.status_code == 200:
            return self._get_required_element_from_response(response.json())
        else:
            return self._handle_response(204, u'delete assets', response)

    @docstring_parameter({'str_type': STR_TYPE_NAME})
    def _delete(self, asset_uid):
        Assets._validate_type(asset_uid, u'asset_uid', STR_TYPE, True)

        response = requests.delete(self._client.service_instance._href_definitions.get_asset_href(asset_uid), params=self._client._params(),
                                   headers=self._client._get_headers())


        # if response.status_code == 200:
        #     return self._get_required_element_from_response(response.json())
        # else:
        #     return self._handle_response(204, u'delete assets', response)


    # @docstring_parameter({'str_type': STR_TYPE_NAME})
    # def get_href(self, asset_uid):
    #     """
    #        Get metadata of stored space(s). If space UID is not specified, it returns all the spaces metadata.
    #
    #        **Parameters**
    #
    #        .. important::
    #             #. **space_uid**: Space UID (optional)\n
    #                **type**: str\n
    #             #. **limit**:  limit number of fetched records (optional)\n
    #                **type**: int\n
    #
    #        **Output**
    #
    #        .. important::
    #             **returns**: metadata of stored space(s)\n
    #             **return type**: dict
    #             dict (if UID is not None) or {"resources": [dict]} (if UID is None)\n
    #
    #        .. note::
    #             If UID is not specified, all spaces metadata is fetched\n
    #
    #        **Example**
    #
    #         >>> space_details = client.spaces.get_details(space_uid)
    #         >>> space_details = client.spaces.get_details()
    #     """
    #
    #
    #     Assets._validate_type(asset_uid, u'asset_uid', STR_TYPE, True)
    #
    #     response = requests.get(self._client.service_instance._href_definitions.get_data_asset_href(asset_uid), params=self._client._params(),
    #                             headers=self._client._get_headers())
    #
    #     if response.status_code == 200:
    #         return response.json()["href"]
    #     else:
    #         return self._handle_response(200, u'spaces assets', response)

    def _get_required_element_from_response(self, response_data):

        WMLResource._validate_type(response_data, u'data assets response', dict)

        import copy
        new_el = {'metadata': copy.copy(response_data['metadata'])}

        try:
            new_el['metadata']['guid'] = response_data['metadata']['asset_id']
            new_el['metadata']['href'] = response_data['href'] if not (self._client.WSD and self._client.default_project_id) else self._client.service_instance._href_definitions.get_base_asset_href(response_data['metadata']['asset_id']) + "?" + "project_id=" + response_data['metadata']['project_id']

            new_el['metadata']['asset_type'] = response_data['metadata']['asset_type']
            new_el['metadata']['created_at'] = response_data['metadata']['created_at']
            new_el['metadata']['last_updated_at'] = response_data['metadata']['usage'].get('last_updated_at')

            if self._client.default_space_id is not None:
                new_el['metadata']['space_id'] = response_data['metadata']['space_id']

            elif self._client.default_project_id is not None:
                new_el['metadata']['project_id'] = response_data['metadata']['project_id']

            if 'entity' in response_data:
                new_el['entity'] = response_data['entity']

            if self._client.ICP_30 is not None or self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                if "attachments" in response_data and response_data[u'attachments']:
                    new_el[u'metadata'].update({'attachment_id': response_data[u'attachments'][0][u'id']})
            if self._client.CLOUD_PLATFORM_SPACES or self._client.ICP_PLATFORM_SPACES:
                href_without_host = response_data['href'].split('.com')[-1]
                new_el[u'metadata'].update({'href':href_without_host})
            return new_el
        except Exception as e:
            raise WMLClientError(Messages.get_message(response_data, message_id="failed_to_read_response_from_down_stream_service"))



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging, re, os

from ibm_watson_machine_learning.libs.repo.mlrepository import MetaNames
from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.runtimes_artifact import RuntimesArtifact
from ibm_watson_machine_learning.libs.repo.swagger_client.api_client import ApiException
import json
from .runtimes_adapter import WmlRuntimesAdapter
from ibm_watson_machine_learning.libs.repo.swagger_client.models import RuntimeSpecDefinitionInput, RuntimeSpecDefinitionInputCustomLibraries,\
    PatchOperationRuntimeSpec, RuntimeSpecDefinitionInputPlatform



logger = logging.getLogger('RuntimesCollection')


class RuntimesCollection:
    """
    Client operating on runtimes in repository service.

    :param str base_path: base url to Watson Machine Learning instance
    :param MLRepositoryApi repository_api: client connecting to repository rest api
    :param MLRepositoryClient client: high level client used for simplification and argument for constructors
    """
    def __init__(self, base_path, repository_api, client):

        from ibm_watson_machine_learning.libs.repo.mlrepositoryclient import MLRepositoryClient
        from ibm_watson_machine_learning.libs.repo.mlrepositoryclient import MLRepositoryApi

        if not isinstance(base_path, str) and not isinstance(base_path, unicode):
            raise ValueError('Invalid type for base_path: {}'.format(base_path.__class__.__name__))

        if not isinstance(repository_api, MLRepositoryApi):
            raise ValueError('Invalid type for repository_api: {}'.format(repository_api.__class__.__name__))

        if not isinstance(client, MLRepositoryClient):
            raise ValueError('Invalid type for client: {}'.format(client.__class__.__name__))

        self.base_path = base_path
        self.repository_api = repository_api
        self.client = client

    def all(self, library_name=None, library_id=None):
        """
        Gets info about all runtimes which belong to this user.

        Not complete information is provided by all(). To get detailed information about experiment use get().

        :return: info about runtimes
        :rtype: list[ExperimentsArtifact]
        """

        all_runtimes = self.repository_api.v3_runtime_spec_list(library_name, library_id)
        list_runtimes_artifact = []
        if all_runtimes is not None:
            resr = all_runtimes.resources
            for iter1 in resr:
                list_runtimes_artifact.append(WmlRuntimesAdapter(iter1, self.client).artifact())
            return list_runtimes_artifact
        else:
            return []

    def get(self, runtimes_id):

        """
        Gets detailed information about runtimes.

        :param str runtimes_id: uid used to identify experiment
        :return: returned object has all attributes of SparkPipelineArtifact but its class name is PipelineArtifact
        :rtype: PipelineArtifact(SparkPipelineLoader)
        """

        if not isinstance(runtimes_id, str) and not isinstance(runtimes_id, unicode):
            raise ValueError('Invalid type for experiment_id: {}'.format(runtimes_id.__class__.__name__))
        if(runtimes_id.__contains__("/v4/runtimes")):
            matched = re.search('.*/v4/runtimes/([A-Za-z0-9\-]+)', runtimes_id)
            if matched is not None:
                runtimes_id = matched.group(1)
                return self.get(runtimes_id)
            else:
                raise ValueError('Unexpected artifact href: {} format'.format(runtimes_id))
        else:
            runtimes_output = self.repository_api.v3_runtime_spec_get(runtimes_id)
            if runtimes_output is not None:
                return WmlRuntimesAdapter(runtimes_output, self.client).artifact()
            else:
                raise Exception('Library not found'.format(runtimes))

    def remove(self, runtimes_id):
        """
        Removes runtimes with given runtimes_id.

        :param str runtimes_id: uid used to identify experiment
        """

        if not isinstance(runtimes_id, str) and not isinstance(runtimes_id, unicode):
            raise ValueError('Invalid type for runtimes_id: {}'.format(runtimes_id.__class__.__name__))

        if(runtimes_id.__contains__("/v4/runtimes")):
            matched = re.search('.*/v4/runtimes/([A-Za-z0-9\-]+)', runtimes_id)
            if matched is not None:
                runtimes_id_value = matched.group(1)
                self.remove(runtimes_id_value)
            else:
                raise ValueError('Unexpected experiment artifact href: {} format'.format(runtimes_id))
        else:
            return self.repository_api.v3_runtime_spec_delete(runtimes_id)

    def patch(self, runtimes_id, artifact):
        runtimes_patch_input = self.prepare_runtimes_patch_input(artifact)
        runtimes_patch_output = self.repository_api.v3_runtime_spec_patch_with_http_info(runtimes_patch_input, runtimes_id)
        statuscode = runtimes_patch_output[1]

        if statuscode != 200:
            logger.info('Error while patching runtimes: no location header')
            raise ApiException(statuscode, "Error while patching runtimes")

        if runtimes_patch_output is not None:
            new_artifact = WmlRuntimesAdapter(runtimes_patch_output[0], self.client).artifact()
        return new_artifact

    def save(self, artifact, runtimespec_path  = None):
        """
        Saves runtimes in repository service.

        :param artifact : RuntimesArtifact to be saved in the repository service
        :param runtimes : runtimes file path to be saved in the repository service
        :return: saved artifact with changed MetaProps
        :rtype: RuntimesArtifact
        """
        logger.debug('Creating a new WML Runtimes : {}'.format(artifact.name))

        if not issubclass(type(artifact), RuntimesArtifact):
            raise ValueError('Invalid type for artifact: {}'.format(artifact.__class__.__name__))

        runtimes_input = self._prepare_wml_runtimes_input(artifact)
        runtimes_output = self.repository_api.v3_runtime_spec_create_with_http_info(runtimes_input)

        runtimes_id = runtimes_output[0].metadata.guid

        if runtimespec_path is not None:
            if not os.path.exists(runtimespec_path):
                raise IOError('The runtimes specified ( {} ) does not exist.'.format(runtimespec_path))
            artifact.runtimespec_path = runtimespec_path

        if runtimespec_path is None and artifact.runtimespec_path is not None:
            if not os.path.exists(artifact.runtimespec_path):
                raise IOError('The artifact specified ( {} ) does not exist.'.format(artifact.runtimespec_path))

        statuscode = runtimes_output[1]
        if statuscode != 201:
            logger.info('Error while creating runtimes: no location header')
            raise ApiException(statuscode, 'No artifact location')

        if runtimes_output is not None:
            new_artifact = WmlRuntimesAdapter(runtimes_output[0], self.client).artifact()
            new_artifact.runtimespec_path = artifact.runtimespec_path
            if artifact.runtimespec_path is not None:
                self._upload_runtimes_content(new_artifact, artifact.runtimespec_path)
        return new_artifact


    @staticmethod
    def _prepare_wml_runtimes_input(artifact):
        name = None
        version = None
        platform =None
        platform_version = None
        description = None

        name = artifact.meta.prop(MetaNames.RUNTIMES.NAME)
        description = artifact.meta.prop(MetaNames.RUNTIMES.DESCRIPTION)
        platform = json.loads(artifact.meta.prop(MetaNames.RUNTIMES.PLATFORM))
        platform_input = RuntimeSpecDefinitionInputPlatform(platform.get('name'), platform.get('version'))
        custom_libs = json.loads(artifact.meta.prop(MetaNames.RUNTIMES.CUSTOM_LIBRARIES_URLS))
        custom_library_param_list = []
        if custom_libs is not None:
            custom_library_param_list = RuntimeSpecDefinitionInputCustomLibraries(custom_libs.get('urls', None))
        if description is not None and not isinstance(description, str):
            raise ValueError('Invalid data format for description.')
        runtimes_input = RuntimeSpecDefinitionInput(
            name, description, platform_input, custom_library_param_list)
        return runtimes_input

    @staticmethod
    def prepare_runtimes_patch_input(artifact):
        patch_list = []
        patch_input = artifact.meta.prop(MetaNames.RUNTIMES.PATCH_INPUT)
        if isinstance(patch_input, str):
            patch_input_list = json.loads(artifact.meta.prop(MetaNames.RUNTIMES.PATCH_INPUT))
            if isinstance(patch_input_list, list):
                for iter1 in patch_input_list:
                    runtimes_patch = PatchOperationRuntimeSpec(
                        op = iter1.get('op'),
                        path= iter1.get('path'),
                        value = iter1.get('value', None),
                        _from =iter1.get('from', None),
                    )
                    patch_list.append(runtimes_patch)

                return patch_list

    def _upload_runtimes_content(self, runtimes_artifact, artifact_path, query_param=None):
        runtimes_id = runtimes_artifact.uid
        content_stream = runtimes_artifact.reader().read()
        self.repository_api.upload_runtimes(runtimes_id, content_stream)
        content_stream.close()
        runtimes_artifact.reader().close()
        logger.debug('Content uploaded for runtimes created at: {}'.format(runtimes_id))




#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

from ibm_watson_machine_learning.libs.repo.util.library_imports import LibraryChecker
from ibm_watson_machine_learning.libs.repo.base_constants import *

lib_checker = LibraryChecker()

if lib_checker.installed_libs[PYSPARK]:
    from pyspark.ml import Pipeline, PipelineModel

if lib_checker.installed_libs[SCIKIT]:
    from ..mlrepositoryartifact.version_helper import ScikitModelBinary
    try:
        # note only up to scikit version 0.20.3
        from sklearn.externals import joblib
    except ImportError:
        # only for scikit 0.23.*
        import joblib

    if lib_checker.installed_libs[XGBOOST]:
        from xgboost import Booster
        from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.version_helper import XGBoostModelBinary

def get_file_list(dir_path, ext):
    file_list = []
    for file in os.listdir(dir_path):
        if file.endswith(ext):
            file_list.append(os.path.join(dir_path, file))
    return file_list

class SparkPipelineContentLoader(object):
    @staticmethod
    def load_content(content_dir):
        lib_checker.check_lib(PYSPARK)
        return Pipeline.read().load(content_dir)


class MLPipelineContentLoader(object):
    @staticmethod
    def load_content(content_dir):
        lib_checker.check_lib(MLPIPELINE)
        from mlpipelinepy.mlpipeline import MLPipeline
        return MLPipeline.read().load(content_dir)


class IBMSparkPipelineContentLoader(object):
    @staticmethod
    def load_content(content_dir):
        lib_checker.check_lib(IBMSPARKPIPELINE)
        from pipeline import IBMSparkPipeline
        return IBMSparkPipeline.read().load(content_dir)


class SparkPipelineModelContentLoader(object):
    @staticmethod
    def load_content(content_dir):
        lib_checker.check_lib(PYSPARK)
        return PipelineModel.read().load(content_dir)


class MLPipelineModelContentLoader(object):
    @staticmethod
    def load_content(content_dir):
        lib_checker.check_lib(MLPIPELINE)
        from mlpipelinepy.mlpipeline import MLPipelineModel
        return MLPipelineModel.read().load(content_dir)


class IBMSparkPipelineModelContentLoader(object):
    @staticmethod
    def load_content(content_dir):
        lib_checker.check_lib(IBMSPARKPIPELINE)
        from pipeline import IBMSparkPipelineModel
        return IBMSparkPipelineModel.read().load(content_dir)


class ScikitPipelineModelContentLoader(object):
    INCOMPAT_MESSAGE = 'You may be trying to read with python 3 a joblib pickle generated with python 2. This feature is not supported by joblib.'

    @staticmethod
    def load_content(content_dir):
        lib_checker.check_lib(SCIKIT)
        extracted_files = get_file_list(content_dir, ScikitModelBinary.bin_ext())
        if(len(extracted_files)==0):
            extracted_files = get_file_list(content_dir, ScikitModelBinary.bin_ext_v0())
            if (len(extracted_files) == 0) or (extracted_files is None):
                raise Exception("No" + ScikitModelBinary.bin_ext() +
                                " file found in the saved model artifact")
        try:
            full_file_name = extracted_files[0]
            artifact_instance = joblib.load(full_file_name)
            return artifact_instance
        except ValueError as ve_ex:
            if ScikitPipelineModelContentLoader.INCOMPAT_MESSAGE in str(ve_ex):
                raise ValueError("Unable to load the model that was saved in Python 2.x runtime. Retry using the " +
                                 "model saved in Python 3.5 runtime.")
            else:
                raise ve_ex
        except Exception as ex:
            print ('Unable to load scikit model with sklearn.externals.joblib version ' + joblib.__version__)
            print ('Error message: ' + str(ex))
            raise ex



class TensorflowPipelineModelContentLoader(object):

    @staticmethod
    def load_content(content_dir,session,tags):
        lib_checker.check_lib(TENSORFLOW)
        import tensorflow as tf
        from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.version_helper import TensorflowModelBinary

        extracted_files = get_file_list(content_dir, TensorflowModelBinary.bin_ext())
        if len(extracted_files)==0:
            raise Exception("No " + TensorflowModelBinary.bin_ext() +
                            " file found in the saved model artifact")
        try:
            from tensorflow import logging
            logging.set_verbosity(logging.WARN)
            if '2.1.0' in tf.__version__:
                metagraphdef = tf.compat.v1.saved_model.loader.load(session, tags, content_dir)
            else:
                metagraphdef = tf.saved_model.loader.load(session, tags, content_dir)

            sig_def = list(metagraphdef.signature_def.values())[0]
            input_tensors = {}
            output_tensors = {}

            for input_key in list(sig_def.inputs.keys()):
                tensor_name = sig_def.inputs[input_key].name
                input_tensors[input_key] = session.graph.get_tensor_by_name(tensor_name)

            for output_key in list(sig_def.outputs.keys()):
                tensor_name = sig_def.outputs[output_key].name
                output_tensors[output_key] = session.graph.get_tensor_by_name(tensor_name)

            return TensorflowRuntimeArtifact(input_tensors, output_tensors,metagraphdef,session)
        except Exception as ex:
            print ('Unable to load Tensorflow model. Error: ' + str(ex))
            raise ex


class TensorflowRuntimeArtifact(object):
    def __init__(self,
                 input_tensors,
                 output_tensors,
                 metagraphdef,
                 session):
        self.input_tensors = input_tensors
        self.output_tensors = output_tensors
        self.metagraphdef = metagraphdef
        self.session = session


    def get_input_tensors(self):
        return self.input_tensors

    def get_output_tensors(self):
        return self.output_tensors

    def get_metagraphdef(self):
        return self.metagraphdef

    def get_session(self):
        return self.session





#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import print_function
import re
import os
import sys
import pkg_resources
import shutil
import tarfile
import logging
import importlib.util

from typing import Optional
from subprocess import check_call
from packaging import version

import ibm_watson_machine_learning._wrappers.requests as requests

from ibm_watson_machine_learning.wml_client_error import WMLClientError, CannotInstallLibrary

INSTANCE_DETAILS_TYPE = u'instance_details_type'
PIPELINE_DETAILS_TYPE = u'pipeline_details_type'
DEPLOYMENT_DETAILS_TYPE = u'deployment_details_type'
EXPERIMENT_RUN_DETAILS_TYPE = u'experiment_run_details_type'
MODEL_DETAILS_TYPE = u'model_details_type'
DEFINITION_DETAILS_TYPE = u'definition_details_type'
EXPERIMENT_DETAILS_TYPE = u'experiment_details_type'
TRAINING_RUN_DETAILS_TYPE = u'training_run_details_type'
FUNCTION_DETAILS_TYPE = u'function_details_type'
DATA_ASSETS_DETAILS_TYPE = u'data_assets_details_type'
SW_SPEC_DETAILS_TYPE = u'sw_spec_details_type'
HW_SPEC_DETAILS_TYPE = u'hw_spec_details_type'
RUNTIME_SPEC_DETAILS_TYPE = u'runtime_spec_details_type'
LIBRARY_DETAILS_TYPE = u'library_details_type'
SPACES_DETAILS_TYPE = u'spaces_details_type'
MEMBER_DETAILS_TYPE = u'member_details_type'
CONNECTION_DETAILS_TYPE = u'connection_details_type'
PKG_EXTN_DETAILS_TYPE = u'pkg_extn_details_type'
UNKNOWN_ARRAY_TYPE = u'resource_type'
UNKNOWN_TYPE = u'unknown_type'
SPACES_IMPORTS_DETAILS_TYPE = u'spaces_imports_details_type'
SPACES_EXPORTS_DETAILS_TYPE = u'spaces_exports_details_type'

SPARK_MLLIB = u'mllib'
SPSS_FRAMEWORK = u'spss-modeler'
TENSORFLOW_FRAMEWORK = u'tensorflow'
XGBOOST_FRAMEWORK = u'xgboost'
SCIKIT_LEARN_FRAMEWORK = u'scikit-learn'
PMML_FRAMEWORK = u'pmml'

STR_TYPE = type(u'string or unicode')
STR_TYPE_NAME = STR_TYPE.__name__


def is_python_2():
    return sys.version_info[0] == 2


def str_type_conv(string):
    if is_python_2() and type(string) is str:
        return unicode(string)
    else:
        return string


def meta_props_str_conv(meta_props):
    for key in meta_props:
        if is_python_2() and type(meta_props[key]) is str:
            meta_props[key] = unicode(meta_props[key])


def get_url(url, headers, params=None, isIcp=False):
    import ibm_watson_machine_learning._wrappers.requests as requests

    if isIcp:
        return requests.get(url, headers=headers, params=params)
    else:
        return requests.get(url, headers=headers, params=params)


def print_text_header_h1(title):
    title = str_type_conv(title)
    print(u'\n\n' + (u'#' * len(title)) + u'\n')
    print(title)
    print(u'\n' + (u'#' * len(title)) + u'\n\n')


def print_text_header_h2(title):
    title = str_type_conv(title)
    print(u'\n\n' + (u'-' * len(title)))
    print(title)
    print((u'-' * len(title)) + u'\n\n')


def get_type_of_details(details):
    if 'resources' in details:
        return UNKNOWN_ARRAY_TYPE
    elif details is None:
        raise WMLClientError('Details doesn\'t exist.')
    else:
        try:
            plan = 'plan' in details[u'entity']

            if plan:
                return INSTANCE_DETAILS_TYPE

            if re.search(u'\/wml_instances\/[^\/]+$', details[u'metadata'][u'url']) is not None:
                return INSTANCE_DETAILS_TYPE
        except:
            pass
        try:
            if re.search(u'\/pipelines\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return PIPELINE_DETAILS_TYPE
        except:
            pass
        try:
            if 'href' in details[u'metadata'] and re.search(u'\/deployments\/[^\/]+$', details[u'metadata'][u'href']) is not None \
                    or re.search(u'\/deployments\/[^\/]+$', details[u'metadata'][u'id']) is not None \
                    or u'virtual_deployment_downloads' in details[u'entity'][u'status']:
                return DEPLOYMENT_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/experiments\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return EXPERIMENT_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/trainings\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return TRAINING_RUN_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/models\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return MODEL_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/functions\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return FUNCTION_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/runtimes\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return RUNTIME_SPEC_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/libraries\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return LIBRARY_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/spaces\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return SPACES_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/members\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return MEMBER_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/members\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return MEMBER_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/assets\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return DATA_ASSETS_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/software_specifications\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return SW_SPEC_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/hardware_specifications\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return HW_SPEC_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/package_extension\/[^\/]+$', details[u'entity'][u'package_extension'][u'href']) is not None:
                return PKG_EXTN_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/imports\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return SPACES_IMPORTS_DETAILS_TYPE
        except:
            pass

        try:
            if re.search(u'\/exports\/[^\/]+$', details[u'metadata'][u'href']) is not None:
                return SPACES_EXPORTS_DETAILS_TYPE
        except:
            pass

        return UNKNOWN_TYPE


def pack(directory_path):
    pass


def unpack(filename):
    pass


def load_model_from_directory(framework, directory_path):
    if "mllib" in framework:
     from pyspark.ml import PipelineModel
     return PipelineModel.read().load(directory_path)
    if "spss" in framework:
     pass
    if "tensorflow" in framework:
     pass
    if "scikit" in framework or "xgboost" in framework:
        try:
            try:
                from sklearn.externals import joblib
            except ImportError:
                import joblib
            pkl_files = [x for x in os.listdir(directory_path) if x.endswith('.pkl')]

            if len(pkl_files) < 1:
                raise WMLClientError('No pkl files in directory.')

            model_id = pkl_files[0]
            return joblib.load(os.path.join(directory_path, model_id))
        except Exception as e:
            raise WMLClientError('Cannot load model from pkl file.', e)
    if "pmml" in framework:
     pass
    else:
        raise WMLClientError(u'Invalid framework specified: \'{}\'.'.format(framework))


# def load_model_from_directory(framework, directory_path):
#     if framework == SPARK_MLLIB:
#         from pyspark.ml import PipelineModel
#         return PipelineModel.read().load(directory_path)
#     elif framework == SPSS_FRAMEWORK:
#         pass
#     elif framework == TENSORFLOW_FRAMEWORK:
#         pass
#     elif framework == SCIKIT_LEARN_FRAMEWORK or framework == XGBOOST_FRAMEWORK:
#         from sklearn.externals import joblib
#         model_id = directory_path[directory_path.rfind('/') + 1:] + ".pkl"
#         return joblib.load(os.path.join(directory_path, model_id))
#     elif framework == PMML_MODEL:
#         pass
#     else:
#         raise WMLClientError('Invalid framework specified: \'{}\'.'.format(framework))


def load_model_from_package(framework, directory):
    unpack(directory)
    load_model_from_directory(framework, directory)


def save_model_to_file(model, framework, base_path, filename):
    if filename.find('.') != -1:
        base_name = filename[:filename.find('.') + 1]
        file_extension = filename[filename.find('.'):]
    else:
        base_name = filename
        file_extension = 'tar.gz'

    if framework == SPARK_MLLIB:
        model.write.overwrite.save(os.path.join(base_path, base_name))
    elif framework == SPSS_FRAMEWORK:
        pass
    elif framework == TENSORFLOW_FRAMEWORK:
        pass
    elif framework == XGBOOST_FRAMEWORK:
        pass
    elif framework == SCIKIT_LEARN_FRAMEWORK:
        os.makedirs(os.path.join(base_path, base_name))
        try:
            from sklearn.externals import joblib
        except ImportError:
            import joblib
        joblib.dump(model, os.path.join(base_path, base_name, base_name + ".pkl"))
    elif framework == PMML_FRAMEWORK:
        pass
    else:
        raise WMLClientError(u'Invalid framework specified: \'{}\'.'.format(framework))


def format_metrics(latest_metrics_list):
    formatted_metrics = u''

    for i in latest_metrics_list:

        values = i[u'values']

        if len(values) > 0:
            sorted_values = sorted(values, key=lambda k: k[u'name'])
        else:
            sorted_values = values

        for j in sorted_values:
            formatted_metrics = formatted_metrics + i[u'phase'] + ':' + j[u'name']+'='+'{0:.4f}'.format(j[u'value']) + '\n'

    return formatted_metrics


def docstring_parameter(args):
    def dec(obj):
        #obj.__doc__ = obj.__doc__.format(**args)
        return obj
    return dec


def group_metrics(metrics):
    grouped_metrics = []

    if len(metrics) > 0:
        import collections
        grouped_metrics = collections.defaultdict(list)
        for d in metrics:
            k = d[u'phase']
            grouped_metrics[k].append(d)

    return grouped_metrics


class StatusLogger:
    def __init__(self, initial_state):
        self.last_state = initial_state
        print(initial_state, end='')

    def log_state(self, state):
        if state == self.last_state:
            print('.', end='')
        else:
            print('\n{}'.format(state), end='')
            self.last_state = state

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


def get_file_from_cos(cos_credentials):
    import ibm_boto3
    from ibm_botocore.client import Config

    client_cos = ibm_boto3.client(service_name='s3',
        ibm_api_key_id=cos_credentials['IBM_API_KEY_ID'],
        ibm_auth_endpoint=cos_credentials['IBM_AUTH_ENDPOINT'],
        config=Config(signature_version='oauth'),
        endpoint_url=cos_credentials['ENDPOINT'])

    streaming_body = client_cos.get_object(Bucket=cos_credentials['BUCKET'], Key=cos_credentials['FILE'])['Body']
    training_definition_bytes = streaming_body.read()
    streaming_body.close()
    filename = cos_credentials['FILE']
    f = open(filename, 'wb')
    f.write(training_definition_bytes)
    f.close()

    return filename


def extract_model_from_repository(model_uid, client):
    """
        Downloads and extracts archived model from wml repository.
        :param model_uid:
        :param client:
        :return: extracted directory path
    """
    create_empty_directory(model_uid)
    current_dir = os.getcwd()

    os.chdir(model_uid)
    model_dir = os.getcwd()

    fname = 'downloaded_' + model_uid + '.tar.gz'
    client.repository.download(model_uid, filename=fname)

    if fname.endswith("tar.gz"):
        tar = tarfile.open(fname)
        tar.extractall()
        tar.close()
    else:
        raise WMLClientError('Invalid type. Expected tar.gz')

    os.chdir(current_dir)
    return model_dir


def extract_mlmodel_from_archive(archive_path, model_uid):
    """
        Extracts archived model under model uid directory.
        :param model_uid:
        :param archive_path:
        :return: extracted directory path
    """
    create_empty_directory(model_uid)
    current_dir = os.getcwd()

    os.rename(archive_path, os.path.join(model_uid, archive_path))

    os.chdir(model_uid)
    model_dir = os.getcwd()

    if archive_path.endswith("tar.gz"):
        tar = tarfile.open(archive_path)
        tar.extractall()
        tar.close()
    else:
        raise WMLClientError('Invalid type. Expected tar.gz')

    os.chdir(current_dir)
    return os.path.join(model_uid, 'model.mlmodel')


def get_model_filename(directory, model_extension):
    logger = logging.getLogger(__name__)
    model_filepath = None

    for file in os.listdir(directory):
        if file.endswith(model_extension):
            if model_filepath is None:
                model_filepath = os.path.join(directory, file)
            else:
                logger.warning('More than one file with extension \'{}\'.'.format(model_extension))

    if model_filepath is None:
        raise WMLClientError('No file with extension \'{}\'.'.format(model_extension))

    return model_filepath


def delete_directory(directory):
    if os.path.exists(directory):
        shutil.rmtree(directory)


def create_empty_directory(directory):
    delete_directory(directory)
    os.makedirs(directory)


def install_package(package):
    import importlib
    try:
        importlib.import_module(package)
    except ImportError:
        import pip
        pip.main(['install', package])


def is_ipython():
    # checks if the code is run in the notebook
    try:
        get_ipython
        return True
    except Exception:
        return False


def create_download_link(file_path, title="Download file."):
    # creates download link for binary files on notebook filesystem (Watson Studio)

    if is_ipython():
        from IPython.display import HTML
        import base64

        filename = os.path.basename(file_path)

        with open(file_path, 'rb') as file:
            b_model = file.read()
        b64 = base64.b64encode(b_model)
        payload = b64.decode()
        html = '<a download="{file_path}" href="data:binary;base64,{payload}" target="_blank">{title}</a>'
        html = html.format(payload=payload, title=title, file_path=filename)

        return HTML(html)


def convert_metadata_to_parameters(meta_data):
    parameters = []

    if meta_data is not None:
        if is_python_2():
            for key, value in meta_data.iteritems():
                parameters.append({'name': str(key), 'value': value})
        else:
            for key, value in meta_data.items():
                parameters.append({'name': str(key), 'value': value})

    return parameters


def is_of_python_basic_type(el):
    if type(el) in [int, float, bool, str]:
        return True
    elif type(el) in [list, tuple]:
        return all([is_of_python_basic_type(t) for t in el])
    elif type(el) is dict:
        if not all(type(k) == str for k in el.keys()):
            return False

        return is_of_python_basic_type(list(el.values()))
    else:
        return False


class NextResourceGenerator:
    """
    Generator class to produce next list of resources from REST API.

    Parameters
    ----------
    wml_client: APIClient, required
        WML Client Instance

    href: str, required
        Href to the resource
    """

    def __init__(self, wml_client: 'APIClient', url: str, href: str, params: dict = None, _all=False) -> None:
        self.wml_client = wml_client
        self.url = url
        self.next_href = href
        self.params = params
        self.all = _all

    def __iter__(self):
        return self

    def __next__(self):
        if self.next_href is not None:
            response = requests.get(
                url=f"{self.url}/{self.next_href}",
                headers=self.wml_client._get_headers(),
                params=self.params if self.params is not None else self.wml_client.params()
            )
            details_json = self.wml_client.training._handle_response(200, "Get next details", response)

            if self.all:
                self.next_href = details_json.get('next', {'href': None})['href']

            else:
                self.next_href = None

            if 'resources' in details_json:
                resources = details_json['resources']

            elif 'metadata' in details_json:
                resources = [details_json]

            else:
                resources = details_json.get('results', [])

            return resources

        else:
            raise StopIteration


class DisableWarningsLogger:
    """
       Class which disables logging warnings (for example for silent handling WMLClientErrors in try except).

        Example:
        ----------
        try:
            with DisableWarningsLogger():
                throw_wml_error()
        except WMLClientError:
            success = False
    """
    def __enter__(self):
        logging.disable(logging.WARNING)

    def __exit__(self, exit_type, exit_value, exit_traceback):
        logging.disable(logging.NOTSET)


def is_lib_installed(lib_name: str, minimum_version: Optional[str] = None, install: Optional[bool] = False) -> bool:
    """Check if provided library is installed on user environment. If not, tries to install it.

    Parameters
    ----------
    lib_name: str, required
        Library name to check.

    minimum_version: str, optional
        Minimum version of library to check. Default None, we can check if library is installed in overall.

    install: bool, optional
        Indicates to install missing or to low version library.

    Returns
    -------
    Bool (true is library is installed, False otherwise)
    """
    if lib_name in sys.modules:
        installed = True

    elif importlib.util.find_spec(lib_name) is not None:
        installed = True

    else:
        installed = False

    if installed:
        try:
            installed_module_version = fallback_to_pip_for_version_check(lib_name=lib_name)

        except Exception:
            installed_module_version = pkg_resources.get_distribution(lib_name).version

        if minimum_version is not None:
            if version.parse(installed_module_version) < version.parse(minimum_version):
                if install:
                    install_library(lib_name=lib_name, version=minimum_version, strict=False)

    else:
        if install:
            install_library(lib_name=lib_name, version=minimum_version, strict=False)
            installed = True

    return installed


def install_library(lib_name: str, version: Optional[str] = None, strict: Optional[bool] = False) -> None:
    """Try to install library.

    Parameters
    ----------
    lib_name: str, required
        Library name to install.

    version: str, optional
        Version of the library to install.

    strict: bool, optional
        Indicates if we want to install specific version or higher version if available. Default False.
    """
    try:
        if version is not None:
            check_call([sys.executable, "-m", "pip", "install", f"{lib_name}{'==' if strict else '>='}{version}"])

        else:
            check_call([sys.executable, "-m", "pip", "install", lib_name])

    except Exception as e:
        raise CannotInstallLibrary(lib_name=lib_name, reason=str(e))


def fallback_to_pip_for_version_check(lib_name: str) -> str:
    """Use only when you need to check package version by package name with pip."""
    try:
        from pip._internal.utils.misc import get_installed_distributions
        packages = get_installed_distributions()
        installed_module_version = None

        for pkg in packages:
            if pkg.key == lib_name:
                installed_module_version = pkg._version
                break

        if installed_module_version is None:
            raise pkg_resources.DistributionNotFound()

        else:
            return installed_module_version

    except Exception:
        raise pkg_resources.DistributionNotFound()


def prepare_interaction_props_for_cos(source_params: dict, file_name: str) -> dict:
    """If user specified properties for dataset as sheet_name, delimiter etc. we need to
    pass them as interaction properties for Flight Service.

    Parameters
    ----------
    source_params: dictionary, required
        Data source parameters describe data (eg. excel_sheet, encoding etc.)

    file_name: str, required
        Name of the file to download, should consist of file extension.

    Returns
    -------
    Dictionary with COS interaction properties for Flight Serivce.
    """
    interaction_properties = {}
    file_format = None

    encoding = source_params.get("encoding", None)

    if ".xls" in file_name or ".xlsx" in file_name:
        file_format = "excel"
        if source_params.get("excel_sheet"):
            interaction_properties["sheet_name"] = str(source_params.get("excel_sheet"))

    elif '.csv' in file_name:
        if source_params.get("quote_character"):
            interaction_properties["quote_character"] = str(source_params.get("quote_character"))
        if encoding is not None:
            interaction_properties["encoding"] = encoding

        input_file_separator = source_params.get("input_file_separator", ",")
        if input_file_separator != ",":
            file_format = "delimited"
            interaction_properties["field_delimiter"] = input_file_separator
        else:
            file_format = "csv"

    elif '.parquet' in file_name or '.prq' in file_name:
        file_format = 'parquet'

    if file_format is not None:
        interaction_properties["file_format"] = file_format

    return interaction_properties


def modify_details_for_script_and_shiny(details_from_get: dict) -> dict:
    """
    Add the href and id of and asset to the same position as it is returned from the POST method
    it allows the `get_id`/`get_href` method to work with details returned by GET method.
    Parameters
    ---
    details_from_get: dict, required
        Details of script/shiny app acquired using GET method.
    Returns
    ---
    Dictionary with 'guid' and 'href' key added to 'metadata'.
    """
    try:
        details_from_get[u'metadata'][u'href'] = details_from_get[u'href']
        details_from_get[u'metadata'][u'guid'] = details_from_get[u'metadata'][u'asset_id']
    except KeyError:
        pass

    return details_from_get



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ibm_watson_machine_learning.libs.repo.mlrepositoryartifact.hybrid_artifact_loader import HybridArtifactLoader


class HybridPipelineModelLoader(HybridArtifactLoader):
    """
        Returns pipeline model instance associated with this model artifact.

        :return: model
        :rtype: hybrid.model
        """
    def model_instance(self, artifact='full'):
        """
           :param artifact: query param string referring to "pipeline_model" or "full"
           Currently accepts:
           :return: returns a hybrid model content tar.gz file or pipeline_model.json
         """
        return self.load(artifact)



#  (C) Copyright IBM Corp. 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import asyncio
import threading
import websockets
from _collections import deque
import signal
import ssl
import time
import concurrent.futures._base

from ibmfl.connection.connection import ConnectionStatus
from ibmfl.connection.connection import FLConnection, FLSender, FLReceiver
from ibmfl.message.message import Message, ResponseMessage
from ibmfl.message.message_type import MessageType
from ibmfl.message.serializer_types import SerializerTypes
from ibmfl.message.serializer_factory import SerializerFactory

logger = logging.getLogger(__name__)

#TODO: remove  self.temp_rts_id = '4a666f4a-87c0-475d-bfb2-312d186163b4' 

party_rbuffer = deque([])
party_sbuffer = deque([])
pRecvEvt = threading.Event()
pRecvEvt.clear()
pSendEvt = threading.Event()
pSendEvt.clear()

class SWSConnection(FLConnection):

    DEFAULT_HOST = '127.0.0.1'
    DEFAULT_PORT = 8765
    TEMP_rts_id = '4a666f4a-87c0-475d-bfb2-312d186163b4' 
    DEFAULT_CONFIG = {'ip': DEFAULT_HOST, 'port': DEFAULT_PORT, 'rts_id': TEMP_rts_id}
    rts_id = None



    def __init__(self, config):
        self.stopped = False   # flag has been exposed in generated script.
        self.settings = self.process_config(config)
        self.rts_id = self.settings.get('id') 
        logger.debug("METHOD = SWSConnection.init")
        logger.debug(self.settings)


    def initialize(self, **kwargs):
        pass

    def initialize_receiver(self, **kwargs):
        self.router = kwargs.get('router')
        self.settings['verify'] = kwargs.get('verify') if kwargs.get('verify') is not None else True
        self.settings['token'] = kwargs.get('token')
        #TODO: Check if sender is intialized, otherwise ?
        self.receiver = SWSReceiver(self.router, self.sender, self.rts_id, self.settings)
        self.receiver.initialize()
        self.receiver.stoppedStateObj = self

    def initialize_sender(self, **kwargs):
        
        self.sender = SWSSender(self.rts_id, self.settings)
        self.sender.initialize()

    def start(self, **kwargs):
        self.receiver.start()


    def stop(self, **kwargs):
        if self.receiver.stop is not None:
            self.receiver.stop.set_result
            self.stopped = True

    def get_connection_config(self):

        return self.settings


    def process_config(self, config): 
        logger.debug("METHOD = SWSConnection.process_config")
        logger.debug(config)

        if config.get('ip', None) is not None and config.get('port', None) is not None:
            return config
        else:
            return self.DEFAULT_CONFIG


class SWSReceiver(FLReceiver):

    def __init__(self, router, sender, party_id, settings):
        self.router = router
        self.ssl_context = True
        self.sender = sender
        self.stopped = False
        self.stopping = False
        self.party = party_id
        self.settings = settings
        self.stop = None
        self.messageInTransit = None
        self.stoppedStateObj = None

    def initialize(self, **kwargs):
        pass

    def debug_websocket(self, websocket):
        #pass
        logger.debug("*************")
        logger.debug(websocket.remote_address)
        logger.debug(websocket.request_headers.get("rtsid"))
        logger.debug("*************")

    def prepare_message_data_structures(self):
            # party_rbuffer = deque([])
            # party_sbuffer = deque([])
            # pRecvEvt = threading.Event()
            # pRecvEvt.clear()
            # pSendEvt = threading.Event()
            # pSendEvt.clear()
        pass
    

    def start(self, **kwargs):
        logger.debug("METHOD = SWSReceiver.start")
        self.loop = asyncio.new_event_loop()
        self.stop = self.loop.create_future()
        try:
            self.loop.add_signal_handler(signal.SIGTERM, self.stop.set_result, None)
        except NotImplementedError:
            logger.debug("Signal handling is not implemented")
        t1 = threading.Thread(target=self.setup_websocket_connection, args=(self.loop, self.stop), daemon=True)
        t1.start()
       

    def stop(self, **kwargs):
        logger.debug("METHOD = SWSReceiver.stop")
        self.stopped = True
        if self.stoppedStateObj is not None:
            self.stoppedStateObj.stopped = True
        

    def setup_websocket_connection(self, loop, stop):
        logger.debug("METHOD = SWSReceiver.setup_websocket_connection")
        try:
            loop.run_until_complete(self.establish_websocket_connection(loop, stop))

            #asyncio.get_event_loop().run_forever()
        except websockets.exceptions.ConnectionClosedError as ex:
            if self.stopped != True:
                logger.error(
                    "PartySendLoop : Connection closed abnormally. Exception is " + str(ex))
                logger.error("Retrying connection to aggregator")
        except websockets.exceptions.ConnectionClosedOK as ex:
            if self.stopped != True:
                logger.error(
                    "PartySendLoop : Connection closed unexpectedly. Exception is " + str(ex))
                logger.error("Retrying connection to aggregator")
            else:
                logger.debug("Connection Closed OK")
        except websockets.exceptions.WebSocketException as ex:
            logger.error("PartySendLoop : WebsocketException. Details: " + str(ex))
        except Exception as ex:
            logger.error("PartySendLoop : Exception is " + str(ex))
        
        self.stop.set_result


    async def establish_websocket_connection(self, loop, stop):
        logger.debug("METHOD = SWSReceiver.establish_websocket_connection")
        uri = "wss://" + self.settings.get("ip") + ":" + str(self.settings.get("port"))
        logger.debug("SWSReceiver.establish_websocket_connection uri = "+uri)

        headers = { 'rtsid': self.party }
        token = self.settings.get('token')
        if ( token is not None ):
            headers['Authorization'] = token
        if ( not self.settings.get('verify') ) :
            self.ssl_context = ssl.SSLContext()
        logger.debug("SWSReceiver.establish_websocket_connection ssl = "+str(self.ssl_context))
        apikey_header = None
        while not self.stopped:
            try:
                logger.debug("SWSReceiver.establish_websocket_connection ATTEMPT async websockets.connect")
                async with websockets.connect(uri, ping_interval=20, ping_timeout=None, max_size=2 ** 29, read_limit=2 ** 29, write_limit=2**29, ssl=self.ssl_context, loop=loop, extra_headers=headers) as websocket:
                    logger.info("Received Heartbeat from Aggregator")
                    apikey_header = websocket.response_headers.get("apikey", None)
                    if apikey_header is not None:
                        logger.debug("API KEY received from Aggregator, for reconnection authorization")
                        headers["Authorization"] = apikey_header
                    await self.websocket_message_loop(websocket)
            except websockets.exceptions.ConnectionClosedError as ex:
                logger.info("ConnectionClosedError encountered.")
                logger.error(str(ex))

                if self.stopped != True:
                    logger.error(
                        "PartySendLoop : Connection closed abnormally. Exception is " + str(ex))
                    logger.info("Retrying connection to aggregator")
                    continue
            except websockets.exceptions.ConnectionClosedOK as ex:
                logger.info("ConnectionClosedOK encountered.")
                logger.error(str(ex))
                if self.stopped != True:
                    logger.error(
                        "PartySendLoop : Connection closed unexpectedly. Exception is " + str(ex))
                    logger.info("Retrying connection to aggregator")
                    continue
                else:
                    logger.info("Connection Closed OK")
                    break
            except websockets.exceptions.WebSocketException as ex:
                logger.info("WebsocketException encountered.")
                logger.error(str(ex))
                self.stopped = True
                continue
            except ssl.SSLError as ex:
                logger.info("SSL Error encountered.")
                pRecvEvt.set()
                logger.error(str(ex))
                raise Exception("No valid certificate detected. Please replace the default certificate with your own "
                                  "TLS certificate, or set verify to False at your own risk. For more details, please see "
                                  "https://www.ibm.com/docs/en/cloud-paks/cp-data/4.0?topic=client-using-custom-tls-certificate-connect-platform")

            except Exception as ex:
                logger.info("Exception encountered.")
                logger.error(str(ex))
                break

    async def websocket_message_loop(self, websocket):
        logger.debug("METHOD = SWSReceiver.websocket_message_loop")
        self.debug_websocket(websocket)

        receive_message_task = asyncio.ensure_future(self.receive_message_handler(websocket) )
        send_message_task = asyncio.ensure_future(self.send_message_handler(websocket) )
        stop_loop_tasks = False
        while not stop_loop_tasks:
            try:

                done, pending = await asyncio.wait([receive_message_task, send_message_task], return_when=asyncio.FIRST_COMPLETED)

                logger.debug("*******************")
                logger.debug(done)
                logger.debug(pending)
                logger.debug("*******************")

                for task in pending:
                    task.cancel()
                for future in done:
                    if not future.cancelled():
                        logger.debug("Future Processed successfully")
                if len(pending) == 0 :
                    stop_loop_tasks = True
            except websockets.exceptions.ConnectionClosedError as ex:
                logger.exception("$$ EXCEPTION IN websocket_message_loop - ConnectionClosedError encountered.")
                self.invalidate_message_data_structures(websocket.request_headers.get("rtsid"), True)
                self.stopped = True

            except Exception as ex:
                logger.exception("$$ EXCEPTION IN WEBSOCKET MESSAGE LOOP for " + websocket.request_headers.get("rtsid") +"; Retrying connection")
                print(ex)

    async def receive_message_handler(self, websocket):
        logger.debug("METHOD = SWSReceiver.receive_message_handler")
        try:
            async for message in websocket:
                logger.debug("METHOD = SWSReceiver.receive_message_handler")
                self.debug_websocket(websocket)
                await self.process_received_message(message)
                #logger.info(message)
        except concurrent.futures._base.CancelledError as ex:
            logger.debug("task is being cancelled")
        except websockets.exceptions.ConnectionClosedError as ex:
            logger.info("ConnectionClosedError encountered.")
            if self.messageInTransit is not None:
                logger.info("Adding in Transit message back to queue")
                party_sbuffer.append(self.messageInTransit)
                self.messageInTransit = None
        except Exception as ex:
            print(ex)

    async def send_message_handler(self, websocket):
        while self.stopped is not True:
            logger.debug("METHOD = SWSReceiver.send_message_handler")
            self.debug_websocket(websocket)
            message = await self.process_message_to_send(websocket.request_headers.get("rtsid"))
            try:
                if message is not None:
                    await websocket.send(message)
                    self.messageInTransit = message
                    logger.debug("METHOD = SWSReceiver.send_message_handler Message SENT")
                    if self.stopping == True:
                        self.stop.set_result(True)
                        self.stopped = True
                        if self.stoppedStateObj is not None:
                            self.stoppedStateObj.stopped = True
            except websockets.exceptions.ConnectionClosedError as ex:
                logger.info("ConnectionClosedError encountered")
            except websockets.exceptions.ConnectionClosedOK as ex:
                logger.info("ConnectionClosedOK")
            except websockets.exceptions.WebSocketException as ex:
                logger.info("WebSocketException encountered")
            except Exception as ex:
                logger.info("Exception")
                logger.info(str(ex))

    async def process_message_to_send(self, rts_id):
        while self.stopped is not True:
            # if len(party_sbuffer) == 0:
            #     logger.info("pSendEvt.wait(), pSendEvt.is_set = " + str(pSendEvt.is_set()))
            #     pSendEvt.wait()
            try: 
                msgToBeSent = party_sbuffer.popleft()
                log_message = f"METHOD = SWSReceiver.process_message_to_send {msgToBeSent}"
                log_message = (log_message[:300]+ '...') if len(log_message) > 300 else log_message
                logger.debug(log_message)
                #logger.debug("METHOD = SWSReceiver.process_message_to_send ")

                return msgToBeSent
            except IndexError:
                await asyncio.sleep(1)
                continue
            except Exception as ex:
                logger.exception("$$ EXCEPTION IN process_message_to_send")
                print(ex)
    
    async def process_received_message(self, message):
        log_message = f"METHOD = SWSReceiver.process_received_message {message}"
        log_message = (log_message[:300]+ '...') if len(log_message) > 300 else log_message
        logger.debug(log_message)
        self.messageInTransit = None
        serializer = SerializerFactory(SerializerTypes.JSON_PICKLE).build()
        recv_msg = serializer.deserialize(message)
        message_type = str(recv_msg.message_type)
        logger.debug('METHOD = SWSReceiver.process_received_message type = ' + message_type )

        if (recv_msg.message_type == MessageType.STOP.value):
            handler, kwargs = self.router.get_handler(request_path=message_type)
            if handler is not None:
                res_message = handler(recv_msg)
                
            self.sender.send_message(recv_msg.sender_info, res_message, False)
            self.stopping = True
            return

        # check if error
        if (recv_msg.message_type == MessageType.ERROR_AUTH.value):
            logger.info("Authorization Error. Please verify ip address, training ID, RTS, project permission and token")
            self.stopped = True
            party_rbuffer.append(recv_msg)
            pRecvEvt.set()

            return

        # if recv_msg.sender_info is not None and recv_msg.sender_info == self.party:
        #     logger.info("Received Message is reply to previously sent message")
        #     return

        # msg_data = recv_msg.get_data()
        # if msg_data is not None and msg_data.get('status', None) is not None:
        #     res_message = self.handle_response_message()
        #     logger.info("status = " + msg_data.get('status'))
        #     self.sender.send_message(recv_msg.sender_info, res_message, False)
       


        handler, kwargs = self.router.get_handler(request_path=message_type)

        if handler is None:

            party_rbuffer.append(recv_msg)
            pRecvEvt.set()

        else:

            if recv_msg.sender_info is not None:
                if recv_msg.sender_info == self.party:
                    logger.debug("Received Message is reply to previously sent message")
                else:
                    logger.warning("Received Message is reply for message from "+recv_msg.sender_info)
                return
                
            try:
                res_message = handler(recv_msg)
            except Exception as ex:
                res_message = Message()
                data = {'status': 'error', 'message': str(ex)}
                res_message.set_data(data)


            self.sender.send_message(recv_msg.sender_info, res_message, False)

    def handle_stop_message(self):
        logger.info("Received STOP request from aggregator")
        response_msg = ResponseMessage(message_type=MessageType.STOP.value,
                                        id_request=-1,
                                        data={'ACK': True})
        logger.info("received a STOP request")
        time.sleep(30)
        self.stopped = True
        return response_msg

    def handle_response_message(self):
        logger.debug("Received response from aggregator with status payload")
        response_msg = ResponseMessage(message_type=MessageType.ACK.value,
                                            id_request=88888,
                                            data={'ACK': True})
        return response_msg    


class SWSSender(FLSender):

    def __init__(self, source_info, settings):

        self.settings = settings
        self.source_info = source_info

    def initialize(self, **kwargs):
        pass

    def send_message(self, destination, message, expectResponse=False):
        message.add_sender_info(self.source_info)
        message_type = str(message.message_type)
        logger.debug('SWSSender.send_message message_type = ' + message_type + ' to destination = ' + str(destination))
        if (message.message_type == MessageType.REGISTER.value):
            expectResponse = True
            logger.debug('SWSSender.send_message set expectResponse to True for REGISTER message')
        serializer = SerializerFactory(SerializerTypes.JSON_PICKLE).build()
        message = serializer.serialize(message)
        party_sbuffer.append(message)
        # pSendEvt.set()
        logger.debug("############# Thread switch")
        if expectResponse:
            logger.debug("############# Waiting for response")
            pRecvEvt.wait()
            logger.debug("############# pRecvEvt activated")
            response_message = party_rbuffer.popleft()
            pRecvEvt.clear()
            logger.debug(
                'Received serialized message as response: ' + str(response_message))

            return response_message

    
    def cleanup(self):
        pass
